# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock

from oslo_utils import uuidutils

from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import local_link_connection as \
    hook
from ironic.objects import port
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


_INVENTORY = {
    'interfaces': [{
        'name': 'em1',
        'mac_address': '11:11:11:11:11:11',
        'ipv4_address': '1.1.1.1',
        'lldp': [(0, ''),
                 (1, '04885a92ec5459'),
                 (2, '0545746865726e6574312f3138'),
                 (3, '0078')]
    }]
}


class LocalLinkConnectionTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory = copy.deepcopy(_INVENTORY)
        self.plugin_data = {'all_interfaces': {'em1': {}}}
        self.port = obj_utils.create_test_port(
            self.context, uuid=uuidutils.generate_uuid(), node_id=self.node.id,
            address='11:11:11:11:11:11', local_link_connection={})

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_valid_data(self, mock_get_port, mock_port_save):
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertTrue(mock_port_save.called)
            self.assertEqual({'switch_id': '88:5a:92:ec:54:59',
                              'port_id': 'Ethernet1/18'},
                             self.port.local_link_connection)

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_lldp_none(self, mock_get_port, mock_port_save):
        self.inventory['interfaces'][0]['lldp'] = None
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertFalse(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection, {})

    @mock.patch.object(port.Port, 'save', autospec=True)
    def test_interface_not_in_all_interfaces(self, mock_port_save):
        self.plugin_data['all_interfaces'] = {}
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertFalse(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection, {})

    @mock.patch.object(hook.LOG, 'debug', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    @mock.patch.object(port.Port, 'save', autospec=True)
    def test_no_port_in_ironic(self, mock_port_save, mock_get_port, mock_log):
        mock_get_port.return_value = None
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertFalse(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection, {})
            mock_log.assert_called_once_with(
                'Skipping LLDP processing for interface %s of node %s: '
                'matching port not found in Ironic.',
                self.inventory['interfaces'][0]['mac_address'],
                task.node.uuid)

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_port_local_link_connection_already_exists(self,
                                                       mock_get_port,
                                                       mock_port_save):
        self.port['local_link_connection'] = {'switch_id': '11:11:11:11:11:11',
                                              'port_id': 'Ether'}
        mock_get_port.return_value = self.port

        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertTrue(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection,
                             {'switch_id': '11:11:11:11:11:11',
                              'port_id': 'Ether'})

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(hook.LOG, 'warning', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_invalid_tlv_value_hex_format(self, mock_get_port, mock_log,
                                          mock_port_save):
        self.inventory['interfaces'][0]['lldp'] = [(2, 'weee')]
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            mock_log.assert_called_once_with(
                'TLV value for TLV type %d is not in correct format. Ensure '
                'that the TLV value is in hexadecimal format when sent to '
                'ironic. Node: %s', 2, task.node.uuid)
            self.assertFalse(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection, {})

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_invalid_port_id_subtype(self, mock_get_port, mock_port_save):
        # First byte of TLV value is processed to calculate the subtype for
        # the port ID, Subtype 6 ('06...') isn't a subtype supported by this
        # hook, so we expect it to skip this TLV.
        self.inventory['interfaces'][0]['lldp'][2] = (
            2, '0645746865726e6574312f3138')
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertTrue(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection,
                             {'switch_id': '88:5a:92:ec:54:59'})

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_port_id_subtype_mac(self, mock_get_port, mock_port_save):
        self.inventory['interfaces'][0]['lldp'][2] = (
            2, '03885a92ec5458')
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertTrue(mock_port_save.called)
            self.assertEqual(self.port.local_link_connection,
                             {'port_id': '88:5a:92:ec:54:58',
                              'switch_id': '88:5a:92:ec:54:59'})

    @mock.patch.object(port.Port, 'save', autospec=True)
    @mock.patch.object(port.Port, 'get_by_address', autospec=True)
    def test_invalid_chassis_id_subtype(self, mock_get_port, mock_port_save):
        # First byte of TLV value is processed to calculate the subtype for
        # the chassis ID, Subtype 5 ('05...') isn't a subtype supported by
        # this hook, so we expect it to skip this TLV.
        self.inventory['interfaces'][0]['lldp'][1] = (1, '05885a92ec5459')
        mock_get_port.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.LocalLinkConnectionHook().__call__(task, self.inventory,
                                                    self.plugin_data)
            self.assertTrue(mock_port_save.called)
            self.assertEqual({'port_id': 'Ethernet1/18'},
                             self.port.local_link_connection)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _

opts = [
    cfg.BoolOpt('send_sensor_data',
                default=False,
                deprecated_group='conductor',
                deprecated_name='send_sensor_data',
                help=_('Enable sending sensor data message via the '
                       'notification bus.')),
    cfg.IntOpt('interval',
               default=600,
               min=1,
               deprecated_group='conductor',
               deprecated_name='send_sensor_data_interval',
               help=_('Seconds between conductor sending sensor data message '
                      'via the notification bus. This was originally for '
                      'consumption via ceilometer, but the data may also '
                      'be consumed via a plugin like '
                      'ironic-prometheus-exporter or any other message bus '
                      'data collector.')),
    cfg.IntOpt('workers',
               default=4, min=1,
               deprecated_group='conductor',
               deprecated_name='send_sensor_data_workers',
               help=_('The maximum number of workers that can be started '
                      'simultaneously for send data from sensors periodic '
                      'task.')),
    cfg.IntOpt('wait_timeout',
               default=300,
               deprecated_group='conductor',
               deprecated_name='send_sensor_data_wait_timeout',
               help=_('The time in seconds to wait for send sensors data '
                      'periodic task to be finished before allowing periodic '
                      'call to happen again. Should be less than '
                      'send_sensor_data_interval value.')),
    cfg.ListOpt('data_types',
                default=['ALL'],
                deprecated_group='conductor',
                deprecated_name='send_sensor_data_types',
                help=_('List of comma separated meter types which need to be '
                       'sent to Ceilometer. The default value, "ALL", is a '
                       'special value meaning send all the sensor data. '
                       'This setting only applies to baremetal sensor data '
                       'being processed through the conductor.')),
    cfg.BoolOpt('enable_for_undeployed_nodes',
                default=False,
                deprecated_group='conductor',
                deprecated_name='send_sensor_data_for_undeployed_nodes',
                help=_('The default for sensor data collection is to only '
                       'collect data for machines that are deployed, however '
                       'operators may desire to know if there are failures '
                       'in hardware that is not presently in use. '
                       'When set to true, the conductor will collect sensor '
                       'information from all nodes when sensor data '
                       'collection is enabled via the send_sensor_data '
                       'setting.')),
    cfg.BoolOpt('enable_for_conductor',
                default=True,
                help=_('If to include sensor metric data for the Conductor '
                       'process itself in the message payload for sensor '
                       'data which allows operators to gather instance '
                       'counts of actions and states to better manage '
                       'the deployment.')),
    cfg.BoolOpt('enable_for_nodes',
                default=True,
                help=_('If to transmit any sensor data for any nodes under '
                       'this conductor\'s management. This option supersedes '
                       'the ``send_sensor_data_for_undeployed_nodes`` '
                       'setting.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='sensor_data')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC Driver for remote system management using Dell Remote Access Card.
"""

from oslo_config import cfg

from ironic.drivers.modules.drac import bios
from ironic.drivers.modules.drac import boot
from ironic.drivers.modules.drac import inspect as drac_inspect
from ironic.drivers.modules.drac import management
from ironic.drivers.modules.drac import power
from ironic.drivers.modules.drac import raid
from ironic.drivers.modules.drac import vendor_passthru
from ironic.drivers.modules.redfish import boot as redfish_boot
from ironic.drivers.modules.redfish import inspect as redfish_inspect
from ironic.drivers.modules.redfish import raid as redfish_raid
from ironic.drivers import redfish


CONF = cfg.CONF


class IDRACHardware(redfish.RedfishHardware):
    """integrated Dell Remote Access Controller hardware type"""

    # Required hardware interfaces

    @property
    def supported_boot_interfaces(self):
        """List of supported boot interfaces."""
        inherited = super().supported_boot_interfaces
        # remove the generic redfish one in favor of the Dell specific
        idx = inherited.index(redfish_boot.RedfishVirtualMediaBoot)
        inherited[idx] = boot.DracRedfishVirtualMediaBoot
        return inherited

    @property
    def supported_management_interfaces(self):
        """List of supported management interfaces."""
        return [management.DracRedfishManagement]

    @property
    def supported_power_interfaces(self):
        """List of supported power interfaces."""
        return ([power.DracRedfishPower]
                + super().supported_power_interfaces)

    # Optional hardware interfaces

    @property
    def supported_bios_interfaces(self):
        """List of supported bios interfaces."""
        return ([bios.DracRedfishBIOS]
                + super().supported_bios_interfaces)

    @property
    def supported_inspect_interfaces(self):
        """List of supported inspect interfaces."""
        # Inspector support should have a higher priority than NoInspect
        # if it is enabled by an operator (implying that the service is
        # installed).
        inherited = super().supported_inspect_interfaces
        # remove the generic redfish one in favor of the Dell specific
        idx = inherited.index(redfish_inspect.RedfishInspect)
        inherited[idx] = drac_inspect.DracRedfishInspect
        return inherited

    @property
    def supported_raid_interfaces(self):
        """List of supported raid interfaces."""
        inherited = super().supported_raid_interfaces
        # remove the generic redfish one in favor of the Dell specific
        idx = inherited.index(redfish_raid.RedfishRAID)
        inherited[idx] = raid.DracRedfishRAID
        return inherited

    @property
    def supported_vendor_interfaces(self):
        """List of supported vendor interfaces."""
        return ([vendor_passthru.DracRedfishVendorPassthru]
                + super().supported_vendor_interfaces)



#   Copyright 2025 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Websocket proxy that is compatible with OpenStack Ironic
noVNC consoles. Leverages websockify.py by Joel Martin
"""

import sys

from oslo_config import cfg
from oslo_log import log
import oslo_middleware.cors as cors_middleware

from ironic.common import exception
from ironic.common import service as ironic_service
from ironic.console import novncproxy_service


CONF = cfg.CONF

LOG = log.getLogger(__name__)


def main():

    # register [cors] config options
    cors_middleware.CORS(None, CONF)

    # Parse config file and command line options, then start logging
    ironic_service.prepare_service('ironic_vncproxy', sys.argv)

    if not CONF.vnc.enabled:
        raise exception.ConfigInvalid("To allow this service to start, set "
                                      "[vnc]enabled = True")

    # Build and start the websocket proxy
    launcher = ironic_service.process_launcher()
    server = novncproxy_service.NoVNCProxyService()
    launcher.launch_service(server)
    sys.exit(launcher.wait())


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for iRMC Power Driver
"""

from unittest import mock

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.irmc import boot as irmc_boot
from ironic.drivers.modules.irmc import common as irmc_common
from ironic.drivers.modules.irmc import power as irmc_power
from ironic.drivers.modules.redfish import power as redfish_power
from ironic.drivers.modules.redfish import utils as redfish_util
from ironic.tests.unit.drivers.modules.irmc import test_common


class IRMCPowerInternalMethodsTestCase(test_common.BaseIRMCTest):

    def test__is_expected_power_state(self):
        target_state = states.SOFT_POWER_OFF
        boot_status_value = irmc_power.BOOT_STATUS_VALUE['unknown']
        self.assertTrue(irmc_power._is_expected_power_state(
            target_state, boot_status_value))

        target_state = states.SOFT_POWER_OFF
        boot_status_value = irmc_power.BOOT_STATUS_VALUE['off']
        self.assertTrue(irmc_power._is_expected_power_state(
            target_state, boot_status_value))

        target_state = states.SOFT_REBOOT
        boot_status_value = irmc_power.BOOT_STATUS_VALUE['os-running']
        self.assertTrue(irmc_power._is_expected_power_state(
            target_state, boot_status_value))

        target_state = states.SOFT_POWER_OFF
        boot_status_value = irmc_power.BOOT_STATUS_VALUE['os-running']
        self.assertFalse(irmc_power._is_expected_power_state(
            target_state, boot_status_value))

    @mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
                lambda *args, **kwargs: None)
    @mock.patch('ironic.drivers.modules.irmc.power.snmp.SNMPClient',
                spec_set=True, autospec=True)
    def test__wait_power_state_soft_power_off(self, snmpclient_mock):
        target_state = states.SOFT_POWER_OFF
        self.config(snmp_polling_interval=1, group='irmc')
        self.config(soft_power_off_timeout=3, group='conductor')
        snmpclient_mock.return_value = mock.Mock(
            **{'get.side_effect': [8, 8, 2]})

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._wait_power_state(task, target_state)

            task.node.refresh()
            self.assertIsNone(task.node.last_error)
            self.assertEqual(states.POWER_OFF, task.node.power_state)
            self.assertEqual(states.NOSTATE, task.node.target_power_state)

    @mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
                lambda *args, **kwargs: None)
    @mock.patch('ironic.drivers.modules.irmc.power.snmp.SNMPClient',
                spec_set=True, autospec=True)
    def test__wait_power_state_soft_reboot(self, snmpclient_mock):
        target_state = states.SOFT_REBOOT
        self.config(snmp_polling_interval=1, group='irmc')
        self.config(soft_power_off_timeout=3, group='conductor')
        snmpclient_mock.return_value = mock.Mock(
            **{'get.side_effect': [10, 6, 8]})

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._wait_power_state(task, target_state)

            task.node.refresh()
            self.assertIsNone(task.node.last_error)
            self.assertEqual(states.POWER_ON, task.node.power_state)
            self.assertEqual(states.NOSTATE, task.node.target_power_state)

    @mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
                lambda *args, **kwargs: None)
    @mock.patch('ironic.drivers.modules.irmc.power.snmp.SNMPClient',
                spec_set=True, autospec=True)
    def test__wait_power_state_timeout(self, snmpclient_mock):
        target_state = states.SOFT_POWER_OFF
        self.config(snmp_polling_interval=1, group='irmc')
        self.config(soft_power_off_timeout=2, group='conductor')
        snmpclient_mock.return_value = mock.Mock(
            **{'get.side_effect': [8, 8, 8]})

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.IRMCOperationError,
                              irmc_power._wait_power_state,
                              task,
                              target_state,
                              timeout=None)

            task.node.refresh()
            self.assertIsNotNone(task.node.last_error)
            self.assertEqual(states.ERROR, task.node.power_state)
            self.assertEqual(states.NOSTATE, task.node.target_power_state)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_power_on_ok(
            self,
            attach_boot_iso_if_needed_mock,
            get_irmc_client_mock,
            _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        target_state = states.POWER_ON
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._set_power_state(task, target_state)
            attach_boot_iso_if_needed_mock.assert_called_once_with(task)
        irmc_client.assert_called_once_with(irmc_power.scci.POWER_ON)
        self.assertFalse(_wait_power_state_mock.called)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    def test__set_power_state_power_off_ok(self,
                                           get_irmc_client_mock,
                                           _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        target_state = states.POWER_OFF
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._set_power_state(task, target_state)
        irmc_client.assert_called_once_with(irmc_power.scci.POWER_OFF)
        self.assertFalse(_wait_power_state_mock.called)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_reboot_ok(
            self,
            attach_boot_iso_if_needed_mock,
            get_irmc_client_mock,
            _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        target_state = states.REBOOT
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._set_power_state(task, target_state)
            attach_boot_iso_if_needed_mock.assert_called_once_with(task)
        irmc_client.assert_called_once_with(irmc_power.scci.POWER_RESET)
        self.assertFalse(_wait_power_state_mock.called)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_soft_reboot_ok(
            self,
            attach_boot_iso_if_needed_mock,
            get_irmc_client_mock,
            _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        target_state = states.SOFT_REBOOT
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._set_power_state(task, target_state)
            attach_boot_iso_if_needed_mock.assert_called_once_with(task)
        irmc_client.assert_called_once_with(irmc_power.scci.POWER_SOFT_CYCLE)
        _wait_power_state_mock.assert_has_calls(
            [mock.call(task, states.SOFT_POWER_OFF, timeout=None),
             mock.call(task, states.SOFT_REBOOT, timeout=None)])

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_soft_power_off_ok(self,
                                                attach_boot_iso_if_needed_mock,
                                                get_irmc_client_mock,
                                                _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        target_state = states.SOFT_POWER_OFF
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            irmc_power._set_power_state(task, target_state)
        self.assertFalse(attach_boot_iso_if_needed_mock.called)
        irmc_client.assert_called_once_with(irmc_power.scci.POWER_SOFT_OFF)
        _wait_power_state_mock.assert_called_once_with(task, target_state,
                                                       timeout=None)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_invalid_target_state(
            self,
            attach_boot_iso_if_needed_mock,
            _wait_power_state_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              irmc_power._set_power_state,
                              task,
                              states.ERROR)
            self.assertFalse(attach_boot_iso_if_needed_mock.called)
            self.assertFalse(_wait_power_state_mock.called)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_scci_exception(self,
                                             attach_boot_iso_if_needed_mock,
                                             get_irmc_client_mock,
                                             _wait_power_state_mock):
        irmc_client = get_irmc_client_mock.return_value
        irmc_client.side_effect = Exception()
        irmc_power.scci.SCCIClientError = Exception

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.IRMCOperationError,
                              irmc_power._set_power_state,
                              task,
                              states.POWER_ON)
            attach_boot_iso_if_needed_mock.assert_called_once_with(
                task)
            self.assertFalse(_wait_power_state_mock.called)

    @mock.patch.object(irmc_power, '_wait_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_common, 'get_irmc_client', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_boot, 'attach_boot_iso_if_needed',
                       autospec=True)
    def test__set_power_state_snmp_exception(self,
                                             attach_boot_iso_if_needed_mock,
                                             get_irmc_client_mock,
                                             _wait_power_state_mock):
        target_state = states.SOFT_REBOOT
        _wait_power_state_mock.side_effect = exception.SNMPFailure(
            "fake exception")

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.IRMCOperationError,
                              irmc_power._set_power_state,
                              task,
                              target_state)
            attach_boot_iso_if_needed_mock.assert_called_once_with(
                task)
            get_irmc_client_mock.return_value.assert_called_once_with(
                irmc_power.STATES_MAP[target_state])
            _wait_power_state_mock.assert_called_once_with(
                task, states.SOFT_POWER_OFF, timeout=None)


class IRMCPowerTestCase(test_common.BaseIRMCTest):

    def test_get_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            properties = task.driver.get_properties()
            for prop in irmc_common.COMMON_PROPERTIES:
                self.assertIn(prop, properties)

    @mock.patch.object(redfish_util, 'parse_driver_info', autospec=True)
    @mock.patch.object(irmc_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate_default(self, mock_drvinfo, redfish_parsedr_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.power.validate(task)
            mock_drvinfo.assert_called_once_with(task.node)
            redfish_parsedr_mock.assert_not_called()

    @mock.patch.object(redfish_util, 'parse_driver_info', autospec=True)
    @mock.patch.object(irmc_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate_ipmi(self, mock_drvinfo, redfish_parsedr_mock):
        self.node.set_driver_internal_info('irmc_ipmi_succeed', True)
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.power.validate(task)
            mock_drvinfo.assert_called_once_with(task.node)
            redfish_parsedr_mock.assert_not_called()

    @mock.patch.object(redfish_util, 'parse_driver_info', autospec=True)
    @mock.patch.object(irmc_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate_fail_ipmi(self, mock_drvinfo, redfish_parsedr_mock):
        side_effect = exception.InvalidParameterValue("Invalid Input")
        mock_drvinfo.side_effect = side_effect
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.power.validate,
                              task)
            redfish_parsedr_mock.assert_not_called()

    @mock.patch.object(redfish_util, 'parse_driver_info', autospec=True)
    @mock.patch.object(irmc_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate_redfish(self, mock_drvinfo, redfish_parsedr_mock):
        self.node.set_driver_internal_info('irmc_ipmi_succeed', False)
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.power.validate(task)
            mock_drvinfo.assert_called_once_with(task.node)
            redfish_parsedr_mock.assert_called_once_with(task.node)

    @mock.patch.object(redfish_util, 'parse_driver_info', autospec=True)
    @mock.patch.object(irmc_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate_fail_redfish(self, mock_drvinfo, redfish_parsedr_mock):
        self.node.set_driver_internal_info('irmc_ipmi_succeed', False)
        self.node.save()
        side_effect = exception.InvalidParameterValue("Invalid Input")
        redfish_parsedr_mock.side_effect = side_effect
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.power.validate,
                              task)
            mock_drvinfo.assert_called_once_with(task.node)

    @mock.patch.object(redfish_power.RedfishPower, 'get_power_state',
                       autospec=True)
    @mock.patch('ironic.drivers.modules.irmc.power.ipmitool.IPMIPower',
                spec_set=True, autospec=True)
    def test_get_power_state_default(self, mock_IPMIPower, redfish_getpw_mock):
        ipmi_power = mock_IPMIPower.return_value
        ipmi_power.get_power_state.return_value = states.POWER_ON
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertEqual(states.POWER_ON,
                             task.driver.power.get_power_state(task))
            ipmi_power.get_power_state.assert_called_once_with(task)
            redfish_getpw_mock.assert_not_called()

    @mock.patch.object(redfish_power.RedfishPower, 'get_power_state',
                       autospec=True)
    @mock.patch('ironic.drivers.modules.irmc.power.ipmitool.IPMIPower',
                spec_set=True, autospec=True)
    def test_get_power_state_ipmi(self, mock_IPMIPower, redfish_getpw_mock):
        self.node.set_driver_internal_info('irmc_ipmi_succeed', True)
        self.node.save()
        ipmi_power = mock_IPMIPower.return_value
        ipmi_power.get_power_state.return_value = states.POWER_ON
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertEqual(states.POWER_ON,
                             task.driver.power.get_power_state(task))
            ipmi_power.get_power_state.assert_called_once_with(task)
            redfish_getpw_mock.assert_not_called()

    @mock.patch.object(redfish_power.RedfishPower, 'get_power_state',
                       autospec=True)
    @mock.patch('ironic.drivers.modules.irmc.power.ipmitool.IPMIPower',
                spec_set=True, autospec=True)
    def test_get_power_state_redfish(self, mock_IPMIPower, redfish_getpw_mock):
        self.node.set_driver_internal_info('irmc_ipmi_succeed', False)
        self.node.save()
        ipmipw_instance = mock_IPMIPower()
        ipmipw_instance.get_power_state.side_effect = exception.IPMIFailure
        redfish_getpw_mock.return_value = states.POWER_ON
        irmc_power_inst = irmc_power.IRMCPower()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertEqual(states.POWER_ON,
                             irmc_power_inst.get_power_state(task))
            ipmipw_instance.get_power_state.assert_called()
            redfish_getpw_mock.assert_called_once_with(irmc_power_inst, task)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    def test_set_power_state(self, mock_set_power):
        mock_set_power.return_value = states.POWER_ON
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.power.set_power_state(task, states.POWER_ON)
        mock_set_power.assert_called_once_with(task, states.POWER_ON,
                                               timeout=None)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    def test_set_power_state_timeout(self, mock_set_power):
        mock_set_power.return_value = states.POWER_ON
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.power.set_power_state(task, states.POWER_ON,
                                              timeout=2)
        mock_set_power.assert_called_once_with(task, states.POWER_ON,
                                               timeout=2)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_power.IRMCPower, 'get_power_state', spec_set=True,
                       autospec=True)
    def test_reboot_reboot(self, mock_get_power, mock_set_power):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_power.return_value = states.POWER_ON
            task.driver.power.reboot(task)
            mock_get_power.assert_called_once_with(
                task.driver.power, task)
        mock_set_power.assert_called_once_with(task, states.REBOOT,
                                               timeout=None)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_power.IRMCPower, 'get_power_state', spec_set=True,
                       autospec=True)
    def test_reboot_reboot_timeout(self, mock_get_power, mock_set_power):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_power.return_value = states.POWER_ON
            task.driver.power.reboot(task, timeout=2)
            mock_get_power.assert_called_once_with(
                task.driver.power, task)
        mock_set_power.assert_called_once_with(task, states.REBOOT,
                                               timeout=2)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_power.IRMCPower, 'get_power_state', spec_set=True,
                       autospec=True)
    def test_reboot_power_on(self, mock_get_power, mock_set_power):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_power.return_value = states.POWER_OFF
            task.driver.power.reboot(task)
            mock_get_power.assert_called_once_with(
                task.driver.power, task)
        mock_set_power.assert_called_once_with(task, states.POWER_ON,
                                               timeout=None)

    @mock.patch.object(irmc_power, '_set_power_state', spec_set=True,
                       autospec=True)
    @mock.patch.object(irmc_power.IRMCPower, 'get_power_state', spec_set=True,
                       autospec=True)
    def test_reboot_power_on_timeout(self, mock_get_power, mock_set_power):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            mock_get_power.return_value = states.POWER_OFF
            task.driver.power.reboot(task, timeout=2)
            mock_get_power.assert_called_once_with(
                task.driver.power, task)
        mock_set_power.assert_called_once_with(task, states.POWER_ON,
                                               timeout=2)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from ironic.common import exception
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import ramdisk_error \
    as ramdisk_error_hook
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class RamdiskErrorTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory = {'fake': 'fake-inventory'}
        self.plugin_data = {'error': 'There was an error!'}

    def test_ramdisk_error(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            hook = ramdisk_error_hook.RamdiskErrorHook()
            self.assertRaises(exception.HardwareInspectionFailure,
                              hook.preprocess,
                              task=task, inventory=self.inventory,
                              plugin_data=self.plugin_data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os

from sphinx.application import Sphinx

__version__ = "1.0.0"


# Data model #


class Entity:
    """Represents an entity in the profile."""

    def __init__(self, name, src):
        self.name = name
        self.src = src
        self.purpose = src.get('Purpose', '')
        self.writable = src.get('WriteRequirement') == 'Mandatory'
        self.required = (src.get('ReadRequirement') in ('Mandatory', None)
                         or self.writable)


class ActionParameter(Entity):
    """Represents a parameter in an Action."""

    def __init__(self, name, src):
        super().__init__(name, src)
        self.required_values = src.get('ParameterValues') or []
        self.recommended_values = src.get('RecommendedValues') or []


class Action(Entity):
    """Represents an action on a resource."""

    def __init__(self, name, src):
        super().__init__(name, src)
        self.parameters = {
            name: ActionParameter(name, value)
            for name, value in src.get('Parameters', {}).items()
        }


class Resource(Entity):
    """Represents any resource in the profile.

    Both top-level resources and nested fields are represented by this class
    (but actions are not).
    """

    def __init__(self, name, src):
        super().__init__(name, src)
        self.min_support_values = src.get('MinSupportValues')
        self.properties = {
            name: Resource(name, value)
            for name, value in src.get('PropertyRequirements', {}).items()
        }
        self.actions = {
            name: Action(name, value)
            for name, value in src.get('ActionRequirements', {}).items()
        }
        self.link_to = (src['Values'][0]
                        if src.get('Comparison') == 'LinkToResource'
                        else None)


# Rendering #

LEVELS = {0: '=', 1: '-', 2: '~', 3: '^'}
INDENT = ' ' * 4


class NestedWriter:
    """A writer that is nested with indentations."""

    def __init__(self, dest, level=0):
        self.dest = dest
        self.level = level

    def text(self, text):
        print(INDENT * self.level + text, file=self.dest)

    def para(self, text):
        self.text(text)
        print(file=self.dest)

    def _nested_common(self, res):
        required = " **[required]**" if res.required else ""
        writable = " **[writable]**" if res.writable else ""
        self.text(f"``{res.name}``{required}{writable}")
        nested = NestedWriter(self.dest, self.level + 1)
        if res.purpose:
            nested.para(res.purpose)
        return nested

    def action(self, res):
        nested = self._nested_common(res)
        for prop in res.parameters.values():
            nested.action_parameter(prop)
        print(file=self.dest)

    def action_parameter(self, res):
        self._nested_common(res)
        print(file=self.dest)

    def resource(self, res):
        nested = self._nested_common(res)
        for prop in res.properties.values():
            nested.resource(prop)
        if res.link_to:
            # NOTE(dtantsur): this is a bit hacky, but we don't have
            # definitions for all possible collections.
            split = res.link_to.split('Collection')
            if len(split) > 1:
                nested.text("Link to a collection of "
                            f":ref:`Redfish-{split[0]}` resources.")
            else:
                nested.text(f"Link to a :ref:`Redfish-{res.link_to}` "
                            "resource.")

        print(file=self.dest)


class Writer(NestedWriter):

    def __init__(self, dest):
        super().__init__(dest)

    def title(self, text, level=1):
        print(text, file=self.dest)
        print(LEVELS[level] * len(text), file=self.dest)

    def top_level(self, res):
        required = " **[required]**" if res.required else ""
        self.para(f".. _Redfish-{res.name}:")
        self.title(f"{res.name}")
        self.para(f"{res.purpose}{required}")
        if res.properties:
            self.title("Properties", level=2)
            for name, prop in res.properties.items():
                self.resource(prop)
        if res.actions:
            self.title("Actions", level=2)
            for name, act in res.actions.items():
                self.action(act)


def builder_inited(app: Sphinx):
    source = os.path.join(app.srcdir, app.config.redfish_interop_source)
    with open(source) as fp:
        profile = json.load(fp)
    fname = os.path.basename(source).replace('json', 'rst')
    dstdir = os.path.join(app.srcdir, app.config.redfish_interop_output_dir)
    with open(os.path.join(dstdir, fname), 'wt') as dest:
        w = Writer(dest)
        w.title(f"{profile['ProfileName']} {profile['ProfileVersion']}", 0)
        w.para(profile['Purpose'])

        try:
            for name, value in sorted(
                (name, value)
                for name, value in profile['Resources'].items()
            ):
                w.top_level(Resource(name, value))
        except Exception:
            import traceback
            traceback.print_exc()
            raise


def setup(app: Sphinx):
    app.connect('builder-inited', builder_inited)
    app.add_config_value('redfish_interop_source', None, 'env', [str])
    app.add_config_value('redfish_interop_output_dir', None, 'env', [str])
    return {'version': __version__}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

from testtools import matchers

from ironic.common import exception
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils


class TestAllocationObject(db_base.DbTestCase, obj_utils.SchemasTestMixIn):

    def setUp(self):
        super(TestAllocationObject, self).setUp()
        self.fake_allocation = db_utils.get_test_allocation(name='host1')

    def test_get_by_id(self):
        allocation_id = self.fake_allocation['id']
        with mock.patch.object(self.dbapi, 'get_allocation_by_id',
                               autospec=True) as mock_get_allocation:
            mock_get_allocation.return_value = self.fake_allocation

            allocation = objects.Allocation.get(self.context, allocation_id)

            mock_get_allocation.assert_called_once_with(allocation_id)
            self.assertEqual(self.context, allocation._context)

    def test_get_by_uuid(self):
        uuid = self.fake_allocation['uuid']
        with mock.patch.object(self.dbapi, 'get_allocation_by_uuid',
                               autospec=True) as mock_get_allocation:
            mock_get_allocation.return_value = self.fake_allocation

            allocation = objects.Allocation.get(self.context, uuid)

            mock_get_allocation.assert_called_once_with(uuid)
            self.assertEqual(self.context, allocation._context)

    def test_get_by_name(self):
        name = self.fake_allocation['name']
        with mock.patch.object(self.dbapi, 'get_allocation_by_name',
                               autospec=True) as mock_get_allocation:
            mock_get_allocation.return_value = self.fake_allocation
            allocation = objects.Allocation.get(self.context, name)

            mock_get_allocation.assert_called_once_with(name)
            self.assertEqual(self.context, allocation._context)

    def test_get_bad_id_and_uuid_and_name(self):
        self.assertRaises(exception.InvalidIdentity,
                          objects.Allocation.get,
                          self.context,
                          'not:a_name_or_uuid')

    def test_create(self):
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        with mock.patch.object(self.dbapi, 'create_allocation',
                               autospec=True) as mock_create_allocation:
            mock_create_allocation.return_value = (
                db_utils.get_test_allocation())

            allocation.create()

            args, _kwargs = mock_create_allocation.call_args
            self.assertEqual(objects.Allocation.VERSION, args[0]['version'])

    def test_save(self):
        uuid = self.fake_allocation['uuid']
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        with mock.patch.object(self.dbapi, 'get_allocation_by_uuid',
                               autospec=True) as mock_get_allocation:
            mock_get_allocation.return_value = self.fake_allocation
            with mock.patch.object(self.dbapi, 'update_allocation',
                                   autospec=True) as mock_update_allocation:
                mock_update_allocation.return_value = (
                    db_utils.get_test_allocation(name='newname',
                                                 updated_at=test_time))
                p = objects.Allocation.get_by_uuid(self.context, uuid)
                p.name = 'newname'
                p.save()

                mock_get_allocation.assert_called_once_with(uuid)
                mock_update_allocation.assert_called_once_with(
                    uuid, {'version': objects.Allocation.VERSION,
                           'name': 'newname'})
                self.assertEqual(self.context, p._context)
                res_updated_at = (p.updated_at).replace(tzinfo=None)
                self.assertEqual(test_time, res_updated_at)

    def test_refresh(self):
        uuid = self.fake_allocation['uuid']
        returns = [self.fake_allocation,
                   db_utils.get_test_allocation(name='newname')]
        expected = [mock.call(uuid), mock.call(uuid)]
        with mock.patch.object(self.dbapi, 'get_allocation_by_uuid',
                               side_effect=returns,
                               autospec=True) as mock_get_allocation:
            p = objects.Allocation.get_by_uuid(self.context, uuid)
            self.assertEqual(self.fake_allocation['name'], p.name)
            p.refresh()
            self.assertEqual('newname', p.name)

            self.assertEqual(expected, mock_get_allocation.call_args_list)
            self.assertEqual(self.context, p._context)

    def test_save_after_refresh(self):
        # Ensure that it's possible to do object.save() after object.refresh()
        db_allocation = db_utils.create_test_allocation()
        p = objects.Allocation.get_by_uuid(self.context, db_allocation.uuid)
        p_copy = objects.Allocation.get_by_uuid(self.context,
                                                db_allocation.uuid)
        p.name = 'newname'
        p.save()
        p_copy.refresh()
        p.copy = 'newname2'
        # Ensure this passes and an exception is not generated
        p_copy.save()

    def test_list(self):
        with mock.patch.object(self.dbapi, 'get_allocation_list',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_allocation]
            allocations = objects.Allocation.list(self.context)
            self.assertThat(allocations, matchers.HasLength(1))
            self.assertIsInstance(allocations[0], objects.Allocation)
            self.assertEqual(self.context, allocations[0]._context)

    def test_payload_schemas(self):
        self._check_payload_schemas(objects.allocation,
                                    objects.Allocation.fields)


class TestConvertToVersion(db_base.DbTestCase):

    def setUp(self):
        super(TestConvertToVersion, self).setUp()
        self.fake_allocation = db_utils.get_test_allocation()

    def test_owner_supported_missing(self):
        # Physical network not set, should be set to default.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        delattr(allocation, 'owner')
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.1")
        self.assertIsNone(allocation.owner)
        self.assertEqual({'owner': None}, allocation.obj_get_changes())

    def test_owner_supported_set(self):
        # Physical network set, no change required.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        allocation.owner = 'owner1'
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.1")
        self.assertEqual('owner1', allocation.owner)
        self.assertEqual({}, allocation.obj_get_changes())

    def test_owner_unsupported_missing(self):
        # Physical network not set, no change required.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        delattr(allocation, 'owner')
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.0")
        self.assertNotIn('owner', allocation)
        self.assertEqual({}, allocation.obj_get_changes())

    def test_owner_unsupported_set_remove(self):
        # Physical network set, should be removed.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        allocation.owner = 'owner1'
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.0")
        self.assertNotIn('owner', allocation)
        self.assertEqual({}, allocation.obj_get_changes())

    def test_owner_unsupported_set_no_remove_non_default(self):
        # Physical network set, should be set to default.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        allocation.owner = 'owner1'
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.0", False)
        self.assertIsNone(allocation.owner)
        self.assertEqual({'owner': None}, allocation.obj_get_changes())

    def test_owner_unsupported_set_no_remove_default(self):
        # Physical network set, no change required.
        allocation = objects.Allocation(self.context, **self.fake_allocation)
        allocation.owner = None
        allocation.obj_reset_changes()
        allocation._convert_to_version("1.0", False)
        self.assertIsNone(allocation.owner)
        self.assertEqual({}, allocation.obj_get_changes())



# coding=utf-8

# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2012 NTT DOCOMO, INC.
# Copyright 2014 International Business Machines Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
IPMI power manager driver.

Uses the 'ipmitool' command (http://ipmitool.sourceforge.net/) to remotely
manage hardware.  This includes setting the boot device, getting a
serial-over-LAN console, and controlling the power state of the machine.

NOTE THAT CERTAIN DISTROS MAY INSTALL openipmi BY DEFAULT, INSTEAD OF ipmitool,
WHICH PROVIDES DIFFERENT COMMAND-LINE OPTIONS AND *IS NOT SUPPORTED* BY THIS
DRIVER.
"""

import contextlib
import os
import re
import subprocess
import tempfile
import time

from eventlet.green import subprocess as green_subprocess
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import strutils

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import states
from ironic.common import utils
from ironic.conductor import task_manager
from ironic.conductor import utils as cond_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import boot_mode_utils
from ironic.drivers.modules import console_utils
from ironic.drivers import utils as driver_utils


LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

VALID_PRIV_LEVELS = ['ADMINISTRATOR', 'CALLBACK', 'OPERATOR', 'USER']

VALID_PROTO_VERSIONS = ('2.0', '1.5')

REQUIRED_PROPERTIES = {
    'ipmi_address': _("IP address or hostname of the node. Required.")
}
OPTIONAL_PROPERTIES = {
    'ipmi_password': _("password. Optional."),
    'ipmi_hex_kg_key': _("Kg key for IPMIv2 authentication. "
                         "The key is expected in hexadecimal format. "
                         "Optional."),
    'ipmi_port': _("remote IPMI RMCP port. Optional."),
    'ipmi_priv_level': _("privilege level; default is ADMINISTRATOR. One of "
                         "%s. Optional.") % ', '.join(VALID_PRIV_LEVELS),
    'ipmi_username': _("username; default is NULL user. Optional."),
    'ipmi_bridging': _("bridging_type; default is \"no\". One of \"single\", "
                       "\"dual\", \"no\". Optional."),
    'ipmi_transit_channel': _("transit channel for bridged request. Required "
                              "only if ipmi_bridging is set to \"dual\"."),
    'ipmi_transit_address': _("transit address for bridged request. Required "
                              "only if ipmi_bridging is set to \"dual\"."),
    'ipmi_target_channel': _("destination channel for bridged request. "
                             "Required only if ipmi_bridging is set to "
                             "\"single\" or \"dual\"."),
    'ipmi_target_address': _("destination address for bridged request. "
                             "Required only if ipmi_bridging is set "
                             "to \"single\" or \"dual\"."),
    'ipmi_local_address': _("local IPMB address for bridged requests. "
                            "Used only if ipmi_bridging is set "
                            "to \"single\" or \"dual\". Optional."),
    'ipmi_protocol_version': _('the version of the IPMI protocol; default '
                               'is "2.0". One of "1.5", "2.0". Optional.'),
    'ipmi_force_boot_device': _("Whether Ironic should specify the boot "
                                "device to the BMC each time the server "
                                "is turned on, eg. because the BMC is not "
                                "capable of remembering the selected boot "
                                "device across power cycles; default value "
                                "is False. Optional."),
    'ipmi_disable_boot_timeout': _('By default ironic will send a raw IPMI '
                                   'command to disable the 60 second timeout '
                                   'for booting. Setting this option to '
                                   'False will NOT send that command on '
                                   'this node. The '
                                   '[ipmi]disable_boot_timeout will be '
                                   'used if this option is not set. '
                                   'Optional.'),
    'ipmi_cipher_suite': _('The number of a cipher suite to use. Only 3 '
                           '(AES-128 with SHA1) or 17 (AES-128 with SHA256) '
                           'should ever be used; 17 is preferred. The '
                           'default value depends on the ipmitool version, '
                           'some recent versions have switched from 3 to 17.'),
}
COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)
CONSOLE_PROPERTIES = {
    'ipmi_terminal_port': _("node's UDP port to connect to. Only required for "
                            "console access.")
}
BRIDGING_OPTIONS = [('local_address', '-m'),
                    ('transit_channel', '-B'), ('transit_address', '-T'),
                    ('target_channel', '-b'), ('target_address', '-t')]

LAST_CMD_TIME = {}
TIMING_SUPPORT = None
SINGLE_BRIDGE_SUPPORT = None
DUAL_BRIDGE_SUPPORT = None
TMP_DIR_CHECKED = None

ipmitool_command_options = {
    'timing': ['ipmitool', '-N', '0', '-R', '0', '-h'],
    'single_bridge': ['ipmitool', '-m', '0', '-b', '0', '-t', '0', '-h'],
    'dual_bridge': ['ipmitool', '-m', '0', '-b', '0', '-t', '0',
                    '-B', '0', '-T', '0', '-h']}

# Note(etingof): For more information on IPMI error codes and `ipmitool`
# human interface please refer to:
# https://www.intel.com/content/dam/www/public/us/en/documents/product-briefs/ipmi-second-gen-interface-spec-v2-rev1-1.pdf
# https://github.com/scottjg/ipmitool/blob/master/lib/ipmi_strings.c#L367
#
# Note(TheJulia): The strings below are hardcoded in ipmitool and get
# substituted in return for the error code received from the IPMI controller.
# As of 1.8.15, no internationalization support appears to be in ipmitool
# which means the strings should always be returned in this form regardless
# of locale.
IPMITOOL_RETRYABLE_FAILURES = ['insufficient resources for session',
                               # Generic completion codes considered retryable
                               'Node busy',
                               'Timeout',
                               'Out of space',
                               'BMC initialization in progress']

# NOTE(lucasagomes): A mapping for the boot devices and their hexadecimal
# value. For more information about these values see the "Set System Boot
# Options Command" section 28.13 of the link below (page 392)
# http://www.intel.com/content/www/us/en/servers/ipmi/ipmi-second-gen-interface-spec-v2-rev1-1.html  # noqa
BOOT_DEVICE_HEXA_MAP = {
    boot_devices.PXE: '0x04',
    boot_devices.DISK: '0x08',
    boot_devices.CDROM: '0x14',
    boot_devices.BIOS: '0x18',
    boot_devices.SAFE: '0x0c'
}
# NOTE(TheJulia): Inline notes for ipmi raw boot device commands.
# Per spec (intel 2nd gen ipmi interface, v2, rev1-1)
#  - bit 7 is CMOS clear
#  - bit 6 is Keyboard lock
#  - bit 5-2 is the boot device selector
#    * where 0000b is to do nothing
#    * where 0001b is to force pxe
#    * where 0010b is boot from hard drive
#    * And anything above 1100b is listed reserved
#      which is 0x30 for ipmitool raw commands!
#  - bit 1 is screen blank
#  - bit 0 is lock out reset button
#
# Effectively this results in
# 00000100b == 0x04 for booting from PXE
# 00001000b == 0x08 for booting from the primary hard disk.
# 00100000b == 0x20 is remotely connected virtual media
# 00100100b == 0x24 which is primary remote boot media as set
#                   via the OEM initiator mailbox information.
#                   Like.. iscsi?
#                   However! Supermicro uses this as the UEFI
#                   hard disk. See: https://www.supermicro.com/support/faqs/faq.cfm?faq=25559  # noqa
# 00010000b == 0x30 Enters the reserved territory, and should not
#                   be expected in use.


def _vendor_aware_boot_device_map(task):
    """Returns an altered vendor boot device map based upon vendor."""
    node = task.node
    vendor = node.properties.get('vendor', None)
    boot_dev_map = BOOT_DEVICE_HEXA_MAP.copy()
    boot_mode = boot_mode_utils.get_boot_mode(node)
    if vendor:
        vendor = str(vendor).lower()
        if boot_mode == 'uefi' and vendor == "supermicro":
            # This difference is only known on UEFI mode for supermicro
            # hardware.
            boot_dev_map[boot_devices.DISK] = '0x24'
        # NOTE(TheJulia): Similar differences may exist with Cisco UCS
        # hardware when using IPMI, however at present we don't know
        # what the setting would be.
        # NOTE(TheJulia) I've observed "mc info" manufacturer name of a cisco
        # c-series machine to return "Unknown (0x168B)"
    return boot_dev_map


def _check_option_support(options):
    """Checks if the specific ipmitool options are supported on host.

    This method updates the module-level variables indicating whether
    an option is supported so that it is accessible by any driver
    interface class in this module. It is intended to be called from
    the __init__ method of such classes only.

    :param options: list of ipmitool options to be checked
    :raises: OSError
    """
    for opt in options:
        if _is_option_supported(opt) is None:
            try:
                cmd = ipmitool_command_options[opt]
                # NOTE(cinerama): use subprocess.check_call to
                # check options & suppress ipmitool output to
                # avoid alarming people
                with open(os.devnull, 'wb') as nullfile:
                    subprocess.check_call(cmd, stdout=nullfile,
                                          stderr=nullfile)
            except subprocess.CalledProcessError:
                LOG.info("Option %(opt)s is not supported by ipmitool",
                         {'opt': opt})
                _is_option_supported(opt, False)
            else:
                LOG.info("Option %(opt)s is supported by ipmitool",
                         {'opt': opt})
                _is_option_supported(opt, True)


def _is_option_supported(option, is_supported=None):
    """Indicates whether the particular ipmitool option is supported.

    :param option: specific ipmitool option
    :param is_supported: Optional Boolean. when specified, this value
                         is assigned to the module-level variable indicating
                         whether the option is supported. Used only if a value
                         is not already assigned.
    :returns: True, indicates the option is supported
    :returns: False, indicates the option is not supported
    :returns: None, indicates that it is not aware whether the option
              is supported
    """
    global SINGLE_BRIDGE_SUPPORT
    global DUAL_BRIDGE_SUPPORT
    global TIMING_SUPPORT

    if option == 'single_bridge':
        if (SINGLE_BRIDGE_SUPPORT is None) and (is_supported is not None):
            SINGLE_BRIDGE_SUPPORT = is_supported
        return SINGLE_BRIDGE_SUPPORT
    elif option == 'dual_bridge':
        if (DUAL_BRIDGE_SUPPORT is None) and (is_supported is not None):
            DUAL_BRIDGE_SUPPORT = is_supported
        return DUAL_BRIDGE_SUPPORT
    elif option == 'timing':
        if (TIMING_SUPPORT is None) and (is_supported is not None):
            TIMING_SUPPORT = is_supported
        return TIMING_SUPPORT


def _console_pwfile_path(uuid):
    """Return the file path for storing the ipmi password for a console."""
    file_name = "%(uuid)s.pw" % {'uuid': uuid}
    return os.path.join(CONF.tempdir, file_name)


@contextlib.contextmanager
def _prepare_ipmi_password(driver_info):
    """Prepares the IPMI password by either setting it in the environment

    or creating a temporary file.

    :param driver_info: the ipmitool parameters for accessing a node.
    :returns: the absolute pathname of the password
    :raises: PasswordFileFailedToCreate from creating or writing to the
             temporary file
    """
    if CONF.ipmi.store_cred_in_env:
        yield _persist_ipmi_password(driver_info)
    else:
        f = None
        try:
            f = tempfile.NamedTemporaryFile(mode='w', dir=CONF.tempdir)
            f.write(str(driver_info['password'] or '\0'))
            f.flush()
        except (IOError, OSError) as exc:
            if f is not None:
                f.close()
            raise exception.PasswordFileFailedToCreate(error=exc)
        except Exception:
            with excutils.save_and_reraise_exception():
                if f is not None:
                    f.close()

        try:
            # NOTE(jlvillal): This yield can not be in the try/except block
            # above because an exception by the caller of this function would
            # then get changed to a PasswordFileFailedToCreate exception which
            # would mislead about the problem and its cause.
            yield ('-f', f.name)
        finally:
            if f is not None:
                f.close()


def is_bridging_enabled(node):
    """Check if IPMI bridging is enabled.

    This call is used in the inspector lookup.
    """
    return node.driver_info.get("ipmi_bridging", "no") != "no"


def _parse_driver_info(node):
    """Gets the parameters required for ipmitool to access the node.

    :param node: the Node of interest.
    :returns: dictionary of parameters.
    :raises: InvalidParameterValue when an invalid value is specified
    :raises: MissingParameterValue when a required ipmi parameter is missing.

    """
    info = node.driver_info or {}
    internal_info = node.driver_internal_info or {}
    bridging_types = ['single', 'dual']
    missing_info = [key for key in REQUIRED_PROPERTIES if not info.get(key)]
    if missing_info:
        raise exception.MissingParameterValue(_(
            "Missing the following IPMI credentials in node's"
            " driver_info: %s.") % missing_info)

    address = info.get('ipmi_address')
    username = info.get('ipmi_username')
    password = str(info.get('ipmi_password', ''))
    hex_kg_key = info.get('ipmi_hex_kg_key')
    dest_port = info.get('ipmi_port')
    port = (info.get('ipmi_terminal_port')
            or internal_info.get('allocated_ipmi_terminal_port'))
    priv_level = info.get('ipmi_priv_level', 'ADMINISTRATOR')
    bridging_type = info.get('ipmi_bridging', 'no')
    local_address = info.get('ipmi_local_address')
    transit_channel = info.get('ipmi_transit_channel')
    transit_address = info.get('ipmi_transit_address')
    target_channel = info.get('ipmi_target_channel')
    target_address = info.get('ipmi_target_address')
    protocol_version = str(info.get('ipmi_protocol_version', '2.0'))
    force_boot_device = info.get('ipmi_force_boot_device', False)
    cipher_suite = info.get('ipmi_cipher_suite')

    if not username:
        LOG.warning('ipmi_username is not defined or empty for node %s: '
                    'NULL user will be utilized.', node.uuid)
    if not password:
        LOG.warning('ipmi_password is not defined or empty for node %s: '
                    'NULL password will be utilized.', node.uuid)

    if protocol_version not in VALID_PROTO_VERSIONS:
        valid_versions = ', '.join(VALID_PROTO_VERSIONS)
        raise exception.InvalidParameterValue(_(
            "Invalid IPMI protocol version value %(version)s, the valid "
            "value can be one of %(valid_versions)s") %
            {'version': protocol_version, 'valid_versions': valid_versions})

    if port is not None:
        port = utils.validate_network_port(port, 'ipmi_terminal_port')

    if dest_port is not None:
        dest_port = utils.validate_network_port(dest_port, 'ipmi_port')

    # check if ipmi_bridging has proper value
    if bridging_type == 'no':
        # if bridging is not selected, then set all bridging params to None
        (local_address, transit_channel, transit_address, target_channel,
         target_address) = (None,) * 5
    elif bridging_type in bridging_types:
        # check if the particular bridging option is supported on host
        if not _is_option_supported('%s_bridge' % bridging_type):
            raise exception.InvalidParameterValue(_(
                "Value for ipmi_bridging is provided as %s, but IPMI "
                "bridging is not supported by the IPMI utility installed "
                "on host. Ensure ipmitool version is > 1.8.11"
            ) % bridging_type)

        # ensure that all the required parameters are provided
        params_undefined = [param for param, value in [
            ("ipmi_target_channel", target_channel),
            ('ipmi_target_address', target_address)] if value is None]
        if bridging_type == 'dual':
            params_undefined2 = [param for param, value in [
                ("ipmi_transit_channel", transit_channel),
                ('ipmi_transit_address', transit_address)
            ] if value is None]
            params_undefined.extend(params_undefined2)
        else:
            # if single bridging was selected, set dual bridge params to None
            transit_channel = transit_address = None

        # If the required parameters were not provided,
        # raise an exception
        if params_undefined:
            raise exception.MissingParameterValue(_(
                "%(param)s not provided") % {'param': params_undefined})
    else:
        raise exception.InvalidParameterValue(_(
            "Invalid value for ipmi_bridging: %(bridging_type)s,"
            " the valid value can be one of: %(bridging_types)s"
        ) % {'bridging_type': bridging_type,
             'bridging_types': bridging_types + ['no']})

    if priv_level not in VALID_PRIV_LEVELS:
        valid_priv_lvls = ', '.join(VALID_PRIV_LEVELS)
        raise exception.InvalidParameterValue(_(
            "Invalid privilege level value:%(priv_level)s, the valid value"
            " can be one of %(valid_levels)s") %
            {'priv_level': priv_level, 'valid_levels': valid_priv_lvls})

    if hex_kg_key and len(hex_kg_key) % 2 != 0:
        raise exception.InvalidParameterValue(_(
            "Number of ipmi_hex_kg_key characters is not even"))

    if cipher_suite is not None:
        try:
            int(cipher_suite)
        except (TypeError, ValueError):
            raise exception.InvalidParameterValue(_(
                "Invalid cipher suite %s, expected a number") % cipher_suite)
        if protocol_version == '1.5':
            raise exception.InvalidParameterValue(_(
                "Cipher suites cannot be used with IPMI 1.5"))

    return {
        'address': address,
        'dest_port': dest_port,
        'username': username,
        'password': password,
        'hex_kg_key': hex_kg_key,
        'port': port,
        'uuid': node.uuid,
        'priv_level': priv_level,
        'local_address': local_address,
        'transit_channel': transit_channel,
        'transit_address': transit_address,
        'target_channel': target_channel,
        'target_address': target_address,
        'protocol_version': protocol_version,
        'force_boot_device': force_boot_device,
        'cipher_suite': cipher_suite,
    }


def _get_ipmitool_args(driver_info, pw_file=None):
    ipmi_version = ('lanplus'
                    if driver_info['protocol_version'] == '2.0'
                    else 'lan')

    args = ['ipmitool',
            '-I', ipmi_version,
            '-H', driver_info['address'],
            '-L', driver_info['priv_level']
            ]

    for field, arg in [('dest_port', '-p'),
                       ('username', '-U'),
                       ('hex_kg_key', '-y'),
                       ('cipher_suite', '-C')]:
        if driver_info[field]:
            args.append(arg)
            args.append(driver_info[field])

    for name, option in BRIDGING_OPTIONS:
        if driver_info[name] is not None:
            args.append(option)
            args.append(driver_info[name])

    if pw_file:
        args.append('-f')
        args.append(pw_file)

    if CONF.ipmi.debug:
        args.append('-v')

    # ensure all arguments are strings
    args = [str(arg) for arg in args]

    return args


def _persist_ipmi_password(driver_info):
    """Persists IPMI password for passing to the ipmitool

    :param driver_info: driver info with the ipmitool parameters
    """
    env = {}
    if CONF.ipmi.store_cred_in_env:
        password = driver_info.get('password')
        if password:
            env = {'IPMI_PASSWORD': password}
        return '-E', env

    path = _console_pwfile_path(driver_info['uuid'])
    pw_file = console_utils.make_persistent_password_file(
        path, driver_info['password'] or '\0')

    return '-f', pw_file


def _ipmitool_timing_args():
    if not _is_option_supported('timing'):
        return []

    if not CONF.ipmi.use_ipmitool_retries:
        return [
            '-R', '1',
            '-N', str(CONF.ipmi.min_command_interval)
        ]

    timeout = CONF.ipmi.command_retry_timeout
    interval = CONF.ipmi.min_command_interval
    num_tries = 0
    cmd_duration = 0

    while cmd_duration + interval <= timeout:
        cmd_duration += interval
        # for each attempt, ipmitool adds one second to the previous
        # interval value
        interval += 1
        num_tries += 1

    return [
        '-R', str(num_tries),
        '-N', str(CONF.ipmi.min_command_interval)
    ]


def choose_cipher_suite(actual_cipher_suite):
    """Gives the possible next available cipher suite version.

    Based on CONF.ipmi.cipher_suite_versions and the last cipher suite version
    used that failed. This function is only called if the node doesn't have
    cipher_suite set. Starts using the last element of the list and decreasing
    the index.

    :param actual_cipher_suite: latest cipher suite used in the
        ipmi call.

    :returns: the next possible cipher suite or None in case of empty
        configuration.
    """
    available_cs_versions = CONF.ipmi.cipher_suite_versions
    if not available_cs_versions:
        return None

    if actual_cipher_suite is None:
        return available_cs_versions[-1]
    else:
        try:
            cs_index = available_cs_versions.index(actual_cipher_suite)
        except ValueError:
            return available_cs_versions[-1]

        return available_cs_versions[max(cs_index - 1, 0)]


def check_cipher_suite_errors(cmd_stderr):
    """Checks if the command stderr contains cipher suite errors.

    :param cmd_stderr: The command stderr.

    :returns: True if the cmd_stderr contains a cipher suite error,
        False otherwise.
    """
    cs_errors = [
        "Unsupported cipher suite ID",
        "Error in open session response message : no matching cipher suite",
        "Error in open session response message : invalid role",
    ]
    for cs_err in cs_errors:
        if cmd_stderr is not None and cs_err in cmd_stderr:
            return True
    return False


def update_cipher_suite_cmd(actual_cs, args):
    """Updates variables and the cipher suite cmd.

    This function updates the values for all parameters so they
    can be used in the next retry of _exec_ipmitool.

    :param actual_cs: a string that represents the cipher suite that was
        used in the command.
    :param args: a list that contains the ipmitool command that was executed,
        it will be modified in-place.

    :returns: the next actual_cs
    """
    actual_cs = choose_cipher_suite(actual_cs)
    if '-C' in args:
        cs_index = args.index('-C') + 1
        args[cs_index] = actual_cs
    else:
        args.append('-C')
        args.append(actual_cs)

    return actual_cs


def _exec_ipmitool(driver_info, command, check_exit_code=None,
                   kill_on_timeout=False):
    """Execute the ipmitool command.

    :param driver_info: the ipmitool parameters for accessing a node.
    :param command: the ipmitool command to be executed.
    :param check_exit_code: Single bool, int, or list of allowed exit codes.
    :param kill_on_timeout: if `True`, kill unresponsive ipmitool on
        `min_command_interval` timeout. Default is `False`. Makes no
        effect on Windows.
    :returns: (stdout, stderr) from executing the command.
    :raises: PasswordFileFailedToCreate from creating or writing to the
             temporary file.
    :raises: processutils.ProcessExecutionError from executing the command.

    """
    args = _get_ipmitool_args(driver_info)

    change_cs = (CONF.ipmi.cipher_suite_versions != []
                 and driver_info.get('cipher_suite') is None)
    actual_cs = None

    timeout = CONF.ipmi.command_retry_timeout

    args.extend(_ipmitool_timing_args())

    extra_args = {}

    if kill_on_timeout:
        extra_args['timeout'] = timeout

    if check_exit_code is not None:
        extra_args['check_exit_code'] = check_exit_code

    end_time = (time.time() + timeout)

    num_tries = max((timeout // CONF.ipmi.min_command_interval), 1)
    while True:
        num_tries = num_tries - 1
        # NOTE(tenbrae): ensure that no communications are sent to a BMC more
        #             often than once every min_command_interval seconds.
        time_till_next_poll = CONF.ipmi.min_command_interval - (
            time.time() - LAST_CMD_TIME.get(driver_info['address'], 0))
        if time_till_next_poll > 0:
            time.sleep(time_till_next_poll)
        # Resetting the list that will be utilized so the password arguments
        # from any previous execution are preserved.
        cmd_args = args[:]
        # 'ipmitool' command will prompt password if there is no '-f'
        # option, we set it to '\0' to write a password file to support
        # empty password

        with _prepare_ipmi_password(driver_info) as (flag, env_path):
            cmd_args.append(flag)
            if CONF.ipmi.store_cred_in_env:
                extra_args['env_variables'] = env_path
            else:
                cmd_args.append(env_path)
            cmd_args.extend(command.split(" "))

            try:
                out, err = utils.execute(*cmd_args, **extra_args)
                return out, err
            except processutils.ProcessExecutionError as e:
                if change_cs and check_cipher_suite_errors(e.stderr):
                    actual_cs = update_cipher_suite_cmd(actual_cs, args)
                else:
                    change_cs = False
                with excutils.save_and_reraise_exception() as ctxt:
                    err_list = [
                        x for x in (
                            IPMITOOL_RETRYABLE_FAILURES
                            + CONF.ipmi.additional_retryable_ipmi_errors)
                        if x in str(e)]
                    # If Ironic is doing retries then retry all errors
                    retry_failures = (err_list
                                      or not CONF.ipmi.use_ipmitool_retries)
                    if ((time.time() > end_time)
                        or (num_tries == 0)
                        or not retry_failures):
                        LOG.error('IPMI Error while attempting "%(cmd)s" '
                                  'for node %(node)s. Error: %(error)s',
                                  {'node': driver_info['uuid'],
                                   'cmd': e.cmd, 'error': e})
                    else:
                        ctxt.reraise = False
                        LOG.warning('IPMI Error encountered, retrying '
                                    '"%(cmd)s" for node %(node)s. '
                                    'Error: %(error)s',
                                    {'node': driver_info['uuid'],
                                     'cmd': e.cmd, 'error': e})

            finally:
                LAST_CMD_TIME[driver_info['address']] = time.time()


def _set_and_wait(task, power_action, driver_info, timeout=None):
    """Helper function for performing an IPMI power action

    This method assumes the caller knows the current power state and does not
    check it prior to changing the power state. Most BMCs should be fine, but
    if a driver is concerned, the state should be checked prior to calling this
    method.

    :param power_action: the action Ironic will perform when changing the
      power state of the node.
    :param timeout: timeout (in seconds) positive integer (> 0) for any
      power state. ``None`` indicates to use default timeout.
    :param driver_info: the ipmitool parameters for accessing a node.
    :returns: one of ironic.common.states

    """
    if power_action == states.POWER_ON:
        cmd_name = "on"
        target_state = states.POWER_ON
    elif power_action == states.POWER_OFF:
        cmd_name = "off"
        target_state = states.POWER_OFF
    elif power_action == states.SOFT_POWER_OFF:
        cmd_name = "soft"
        target_state = states.POWER_OFF
        timeout = timeout or CONF.conductor.soft_power_off_timeout
    elif power_action == states.REBOOT:
        cmd_name = "reset"
        target_state = states.POWER_ON

    # NOTE(sambetts): Retries for ipmi power action failure will be handled by
    # the _exec_ipmitool function, so no need to wrap this call in its own
    # retries.
    cmd = "power %s" % cmd_name
    try:
        _exec_ipmitool(driver_info, cmd)
    except (exception.PasswordFileFailedToCreate,
            processutils.ProcessExecutionError,
            subprocess.TimeoutExpired,
            # https://github.com/eventlet/eventlet/issues/624
            green_subprocess.TimeoutExpired) as e:
        LOG.warning("IPMI power action %(cmd)s failed for node %(node_id)s "
                    "with error: %(error)s.",
                    {'node_id': driver_info['uuid'], 'cmd': cmd, 'error': e})
        raise exception.IPMIFailure(cmd=cmd)
    if power_action == states.REBOOT:
        LOG.debug('Sleeping 15 seconds to give the node %s a chance to power '
                  'off before checking its state', task.node.uuid)
        time.sleep(15)
    return cond_utils.node_wait_for_power_state(task, target_state,
                                                timeout=timeout)


def _power_on(task, driver_info, timeout=None):
    """Turn the power ON for this node.

    :param driver_info: the ipmitool parameters for accessing a node.
    :param timeout: the timeout in seconds (> 0) to wait for the power
      action to be completed. ``None`` indicates default timeout".
    :returns: one of ironic.common.states POWER_ON or ERROR.
    :raises: IPMIFailure on an error from ipmitool (from _power_status call).

    """
    return _set_and_wait(task, states.POWER_ON, driver_info, timeout=timeout)


def _power_off(task, driver_info, timeout=None):
    """Turn the power OFF for this node.

    :param driver_info: the ipmitool parameters for accessing a node.
    :param timeout: the timeout in seconds (> 0) to wait for the power
      action to be completed. ``None`` indicates default timeout".
    :returns: one of ironic.common.states POWER_OFF or ERROR.
    :raises: IPMIFailure on an error from ipmitool (from _power_status call).

    """
    return _set_and_wait(task, states.POWER_OFF, driver_info, timeout=timeout)


def _soft_power_off(task, driver_info, timeout=None):
    """Turn the power SOFT OFF for this node.

    :param driver_info: the ipmitool parameters for accessing a node.
    :param timeout: the timeout in seconds (> 0) to wait for the power
      action to be completed. ``None`` indicates default timeout".
    :returns: one of ironic.common.states POWER_OFF or ERROR.
    :raises: IPMIFailure on an error from ipmitool (from _power_status call).

    """
    return _set_and_wait(task, states.SOFT_POWER_OFF, driver_info,
                         timeout=timeout)


def _power_status(driver_info):
    """Get the power status for a node.

    :param driver_info: the ipmitool access parameters for a node.
    :returns: one of ironic.common.states POWER_OFF, POWER_ON or ERROR.
    :raises: IPMIFailure on an error from ipmitool.

    """
    cmd = "power status"
    try:
        out_err = _exec_ipmitool(
            driver_info, cmd, kill_on_timeout=CONF.ipmi.kill_on_timeout)
    except (exception.PasswordFileFailedToCreate,
            processutils.ProcessExecutionError) as e:
        LOG.warning("IPMI power status failed for node %(node_id)s with "
                    "error: %(error)s.",
                    {'node_id': driver_info['uuid'], 'error': e})
        raise exception.IPMIFailure(cmd=cmd)

    if out_err[0] == "Chassis Power is on\n":
        return states.POWER_ON
    elif out_err[0] == "Chassis Power is off\n":
        return states.POWER_OFF
    else:
        return states.ERROR


def _process_sensor(sensor_data):
    sensor_data_fields = sensor_data.split('\n')
    sensor_data_dict = {}
    for field in sensor_data_fields:
        if not field:
            continue
        if field.startswith('<<'):
            # This is debug data, and can be safely ignored for this.
            continue
        kv_value = field.split(':')
        if len(kv_value) != 2:
            continue
        sensor_data_dict[kv_value[0].strip()] = kv_value[1].strip()

    return sensor_data_dict


def _get_sensor_type(node, sensor_data_dict):
    # Have only three sensor type name IDs: 'Sensor Type (Analog)'
    # 'Sensor Type (Discrete)' and 'Sensor Type (Threshold)'

    for key in ('Sensor Type (Analog)', 'Sensor Type (Discrete)',
                'Sensor Type (Threshold)'):
        try:
            return sensor_data_dict[key].split(' ', 1)[0]
        except KeyError:
            continue

    raise exception.FailedToParseSensorData(
        node=node.uuid,
        error=(_("parse ipmi sensor data failed, unknown sensor type"
                 " data: %(sensors_data)s"),
               {'sensors_data': sensor_data_dict}))


def _parse_ipmi_sensors_data(node, sensors_data):
    """Parse the IPMI sensors data and format to the dict grouping by type.

    We run 'ipmitool' command with 'sdr -v' options, which can return sensor
    details in human-readable format, we need to format them to JSON string
    dict-based data for Ceilometer Collector which can be sent it as payload
    out via notification bus and consumed by Ceilometer Collector.

    :param sensors_data: the sensor data returned by ipmitool command.
    :returns: the sensor data with JSON format, grouped by sensor type.
    :raises: FailedToParseSensorData when error encountered during parsing.

    """
    sensors_data_dict = {}
    if not sensors_data:
        return sensors_data_dict

    sensors_data_array = sensors_data.split('\n\n')
    for sensor_data in sensors_data_array:
        sensor_data_dict = _process_sensor(sensor_data)
        if not sensor_data_dict:
            continue

        sensor_type = _get_sensor_type(node, sensor_data_dict)

        # ignore the sensors which has no current 'Sensor Reading' data
        if 'Sensor Reading' in sensor_data_dict:
            sensors_data_dict.setdefault(
                sensor_type,
                {})[sensor_data_dict['Sensor ID']] = sensor_data_dict

    # get nothing, no valid sensor data
    if not sensors_data_dict:
        raise exception.FailedToParseSensorData(
            node=node.uuid,
            error=(_("parse ipmi sensor data failed, get nothing with input"
                     " data: %(sensors_data)s")
                   % {'sensors_data': sensors_data}))
    return sensors_data_dict


@METRICS.timer('send_raw')
@task_manager.require_exclusive_lock
def send_raw(task, raw_bytes):
    """Send raw bytes to the BMC. Bytes should be a string of bytes.

    :param task: a TaskManager instance.
    :param raw_bytes: a string of raw bytes to send, e.g. '0x00 0x01'
    :returns: a tuple with stdout and stderr.
    :raises: IPMIFailure on an error from ipmitool.
    :raises: MissingParameterValue if a required parameter is missing.
    :raises: InvalidParameterValue when an invalid value is specified.

    """
    node_uuid = task.node.uuid
    LOG.debug('Sending node %(node)s raw bytes %(bytes)s',
              {'bytes': raw_bytes, 'node': node_uuid})
    driver_info = _parse_driver_info(task.node)
    cmd = 'raw %s' % raw_bytes

    try:
        out, err = _exec_ipmitool(driver_info, cmd)
        LOG.debug('send raw bytes returned stdout: %(stdout)s, stderr:'
                  ' %(stderr)s', {'stdout': out, 'stderr': err})
    except (exception.PasswordFileFailedToCreate,
            processutils.ProcessExecutionError) as e:
        LOG.exception('IPMI "raw bytes" failed for node %(node_id)s '
                      'with error: %(error)s.',
                      {'node_id': node_uuid, 'error': e})
        raise exception.IPMIFailure(cmd=cmd)

    return out, err


@METRICS.timer('dump_sdr')
def dump_sdr(task, file_path):
    """Dump SDR data to a file.

    :param task: a TaskManager instance.
    :param file_path: the path to SDR dump file.
    :raises: IPMIFailure on an error from ipmitool.
    :raises: MissingParameterValue if a required parameter is missing.
    :raises: InvalidParameterValue when an invalid value is specified.

    """
    node_uuid = task.node.uuid
    LOG.debug('Dump SDR data for node %(node)s to file %(name)s',
              {'name': file_path, 'node': node_uuid})
    driver_info = _parse_driver_info(task.node)
    cmd = 'sdr dump %s' % file_path

    try:
        out, err = _exec_ipmitool(driver_info, cmd)
        LOG.debug('dump SDR returned stdout: %(stdout)s, stderr:'
                  ' %(stderr)s', {'stdout': out, 'stderr': err})
    except (exception.PasswordFileFailedToCreate,
            processutils.ProcessExecutionError) as e:
        LOG.exception('IPMI "sdr dump" failed for node %(node_id)s '
                      'with error: %(error)s.',
                      {'node_id': node_uuid, 'error': e})
        raise exception.IPMIFailure(cmd=cmd)


def _check_temp_dir():
    """Check for Valid temp directory."""
    global TMP_DIR_CHECKED
    # because a temporary file is used to pass the password to ipmitool,
    # we should check the directory
    if TMP_DIR_CHECKED is None:
        try:
            utils.check_dir()
        except (exception.PathNotFound,
                exception.DirectoryNotWritable,
                exception.InsufficientDiskSpace) as e:
            with excutils.save_and_reraise_exception():
                TMP_DIR_CHECKED = False
                err_msg = (_("Ipmitool drivers need to be able to create "
                             "temporary files to pass password to ipmitool. "
                             "Encountered error: %s") % e)
                e.message = err_msg
                LOG.error(err_msg)
        else:
            TMP_DIR_CHECKED = True


def _constructor_checks(driver):
    """Common checks to be performed when instantiating an ipmitool class."""
    try:
        _check_option_support(['timing', 'single_bridge', 'dual_bridge'])
    except OSError:
        raise exception.DriverLoadError(
            driver=driver,
            reason=_("Unable to locate usable ipmitool command in "
                     "the system path when checking ipmitool version"))
    _check_temp_dir()


def _allocate_port(task, host=None):
    node = task.node
    allocated_port = console_utils.acquire_port(host=host)
    node.set_driver_internal_info('allocated_ipmi_terminal_port',
                                  allocated_port)
    node.save()
    return allocated_port


def _release_allocated_port(task):
    node = task.node
    allocated_port = node.del_driver_internal_info(
        'allocated_ipmi_terminal_port')
    if allocated_port:
        node.save()
        console_utils.release_port(allocated_port)


class IPMIPower(base.PowerInterface):

    def __init__(self):
        _constructor_checks(driver=self.__class__.__name__)

    def get_properties(self):
        return COMMON_PROPERTIES

    @METRICS.timer('IPMIPower.validate')
    def validate(self, task):
        """Validate driver_info for ipmitool driver.

        Check that node['driver_info'] contains IPMI credentials.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue if required ipmi parameters are missing.
        :raises: MissingParameterValue if a required parameter is missing.

        """
        _parse_driver_info(task.node)
        # NOTE(tenbrae): don't actually touch the BMC in validate because it is
        #             called too often, and BMCs are too fragile.
        #             This is a temporary measure to mitigate problems while
        #             1314954 and 1314961 are resolved.

    @METRICS.timer('IPMIPower.get_power_state')
    def get_power_state(self, task):
        """Get the current power state of the task's node.

        :param task: a TaskManager instance containing the node to act on.
        :returns: one of ironic.common.states POWER_OFF, POWER_ON or ERROR.
        :raises: InvalidParameterValue if required ipmi parameters are missing.
        :raises: MissingParameterValue if a required parameter is missing.
        :raises: IPMIFailure on an error from ipmitool (from _power_status
            call).

        """
        driver_info = _parse_driver_info(task.node)
        return _power_status(driver_info)

    @METRICS.timer('IPMIPower.set_power_state')
    @task_manager.require_exclusive_lock
    def set_power_state(self, task, power_state, timeout=None):
        """Turn the power on, off, soft reboot, or soft power off.

        :param task: a TaskManager instance containing the node to act on.
        :param power_state: desired power state.
          one of ironic.common.states, POWER_ON, POWER_OFF, SOFT_POWER_OFF,
          or SOFT_REBOOT.
        :param timeout: timeout (in seconds) positive integer (> 0) for any
          power state. The timeout is counted once during power off and once
          during power on for reboots. ``None`` indicates that the default
          timeout will be used.
        :raises: InvalidParameterValue if an invalid power state was specified.
        :raises: MissingParameterValue if required ipmi parameters are missing
        :raises: PowerStateFailure if the power couldn't be set to pstate.

        """
        driver_info = _parse_driver_info(task.node)

        if power_state == states.POWER_ON:
            driver_utils.ensure_next_boot_device(task, driver_info)
            _power_on(task, driver_info, timeout=timeout)
        elif power_state == states.POWER_OFF:
            _power_off(task, driver_info, timeout=timeout)
        elif power_state == states.SOFT_POWER_OFF:
            _soft_power_off(task, driver_info, timeout=timeout)
        elif power_state == states.SOFT_REBOOT:
            if task.node.disable_power_off:
                # There is no way to implement this in ipmitool, apparently
                raise exception.UnsupportedHardwareFeature(
                    node=task.node.uuid,
                    feature=states.SOFT_POWER_OFF)
            _soft_power_off(task, driver_info, timeout=timeout)
            driver_utils.ensure_next_boot_device(task, driver_info)
            _power_on(task, driver_info, timeout=timeout)
        else:
            raise exception.InvalidParameterValue(
                _("set_power_state called "
                  "with invalid power state %s.") % power_state)

    @METRICS.timer('IPMIPower.reboot')
    @task_manager.require_exclusive_lock
    def reboot(self, task, timeout=None):
        """Cycles the power to the task's node.

        :param task: a TaskManager instance containing the node to act on.
        :param timeout: timeout (in seconds) positive integer (> 0) for any
          power state. The timeout is counted once during power off and once
          during power on for reboots. ``None`` indicates that the default
          timeout will be used.
        :raises: MissingParameterValue if required ipmi parameters are missing.
        :raises: InvalidParameterValue if an invalid power state was specified.
        :raises: PowerStateFailure if the final state of the node is not
          POWER_ON or the intermediate state of the node is not POWER_OFF.

        """
        driver_info = _parse_driver_info(task.node)
        if task.node.disable_power_off:
            driver_utils.ensure_next_boot_device(task, driver_info)
            _set_and_wait(task, states.REBOOT, driver_info, timeout=timeout)
            return
        # NOTE(jlvillal): Some BMCs will error if setting power state to off if
        # the node is already turned off.
        current_status = _power_status(driver_info)
        if current_status != states.POWER_OFF:
            _power_off(task, driver_info, timeout=timeout)
        driver_utils.ensure_next_boot_device(task, driver_info)
        _power_on(task, driver_info, timeout=timeout)

    def get_supported_power_states(self, task):
        """Get a list of the supported power states.

        :param task: A TaskManager instance containing the node to act on.
            currently not used.
        :returns: A list with the supported power states defined
                  in :mod:`ironic.common.states`.
        """
        return [states.POWER_ON, states.POWER_OFF, states.REBOOT,
                states.SOFT_REBOOT, states.SOFT_POWER_OFF]


class IPMIManagement(base.ManagementInterface):

    def get_properties(self):
        return COMMON_PROPERTIES

    def __init__(self):
        _constructor_checks(driver=self.__class__.__name__)

    @METRICS.timer('IPMIManagement.validate')
    def validate(self, task):
        """Check that 'driver_info' contains IPMI credentials.

        Validates whether the 'driver_info' property of the supplied
        task's node contains the required credentials information.

        :param task: a task from TaskManager.
        :raises: InvalidParameterValue if required IPMI parameters
            are missing.
        :raises: MissingParameterValue if a required parameter is missing.

        """
        _parse_driver_info(task.node)

    def get_supported_boot_devices(self, task):
        """Get a list of the supported boot devices.

        :param task: a task from TaskManager.
        :returns: A list with the supported boot devices defined
                  in :mod:`ironic.common.boot_devices`.

        """
        return list(BOOT_DEVICE_HEXA_MAP)

    @METRICS.timer('IPMIManagement.set_boot_device')
    @task_manager.require_exclusive_lock
    def set_boot_device(self, task, device, persistent=False):
        """Set the boot device for the task's node.

        Set the boot device to use on next reboot of the node.

        :param task: a task from TaskManager.
        :param device: the boot device, one of
                       :mod:`ironic.common.boot_devices`.
        :param persistent: Boolean value. True if the boot device will
                           persist to all future boots, False if not.
                           Default: False.
        :raises: InvalidParameterValue if an invalid boot device is specified
        :raises: MissingParameterValue if required ipmi parameters are missing.
        :raises: IPMIFailure on an error from ipmitool.

        """
        if device not in self.get_supported_boot_devices(task):
            raise exception.InvalidParameterValue(_(
                "Invalid boot device %s specified.") % device)

        # NOTE(tonyb): Some BMCs do not implement Option 0x03, such as OpenBMC
        # and will error when we try to set this.  Resulting in an abort.  If
        # the BMC doesn't support this timeout there isn't a need to disable
        # it.  Let's use a driver option to signify that.
        # NOTE(kaifeng) [ipmi]disable_boot_timeout provides default value if
        # driver_info/ipmi_disable_boot_timeout is not set.
        idt = task.node.driver_info.get('ipmi_disable_boot_timeout')
        if idt is None:
            idt = CONF.ipmi.disable_boot_timeout
        if strutils.bool_from_string(idt):
            # note(JayF): IPMI spec indicates unless you send these raw bytes
            # the boot device setting times out after 60s. Since it's possible
            # it could be >60s before a node is rebooted, we should always send
            # them.  This mimics pyghmi's current behavior, and the
            # "option=timeout" setting on newer ipmitool binaries.
            timeout_disable = "0x00 0x08 0x03 0x08"
            send_raw(task, timeout_disable)
        else:
            LOG.info('For node %(node_uuid)s, '
                     'driver_info[\'ipmi_disable_boot_timeout\'] is set '
                     'to False, so not sending ipmi boot-timeout-disable',
                     {'node_uuid': task.node.uuid})

        ifbd = task.node.driver_info.get('ipmi_force_boot_device', False)
        if strutils.bool_from_string(ifbd):
            driver_utils.force_persistent_boot(task,
                                               device,
                                               persistent)
            # Reset persistent to False, in case of BMC does not support
            # persistent or we do not have admin rights.
            persistent = False

        # NOTE(lucasagomes): Older versions of the ipmitool utility
        # are not able to set the options "efiboot" and "persistent"
        # at the same time, combining other options seems to work fine,
        # except efiboot. Newer versions of ipmitool (1.8.17) does fix
        # this problem but (some) distros still packaging an older version.
        # To workaround this problem for now we can make use of sending
        # raw bytes to set the boot device for a node in persistent +
        # uefi mode, this will work with newer and older versions of the
        # ipmitool utility. Also see:
        # https://bugs.launchpad.net/ironic/+bug/1611306
        # NOTE(TheJulia): Previously we added raw device support because
        # most distributions are carrying older downstream patched versions
        # of ipmitool where "efiboot" and "persistent" do not work. However,
        # using the embedded support assumes that every vendor out there uses
        # the same device numbers all the time. Reality is not so kind.
        # It should also be noted that the ipmitool chassis bootparm get
        # command also just interprets back the same fields, so if the vendor
        # has set a different value as default, then ipmitool is not going
        # to understand it and may be listing the wrong boot flag as a result.
        # See: https://storyboard.openstack.org/#!/story/2008241
        boot_mode = boot_mode_utils.get_boot_mode(task.node)
        if boot_mode == 'uefi':
            # Long story short: UEFI was added to IPMI after the final spec
            # release occurred. This means BMCs may actually NEED to be
            # explicitly told if the boot is persistent because the
            # BMC may return a value which is explicitly parsed as
            # no change, BUT the BMC may treat that as operational default.
            efi_persistence = '0xe0' if persistent else '0xa0'
            # 0xe0 is persistent UEFI boot, 0xa0 is one-time UEFI boot.
            boot_device_map = _vendor_aware_boot_device_map(task)
            raw_cmd = ('0x00 0x08 0x05 %s %s 0x00 0x00 0x00' %
                       (efi_persistence, boot_device_map[device]))
            send_raw(task, raw_cmd)
            return

        options = []
        if persistent:
            options.append('persistent')
        # NOTE(TheJulia): Once upon a time we had efiboot here. It doesn't
        # work in all cases and directs us to unhappy places if the values
        # are incorrect.
        cmd = "chassis bootdev %s" % device
        if options:
            cmd = cmd + " options=%s" % ','.join(options)
        driver_info = _parse_driver_info(task.node)
        try:
            out, err = _exec_ipmitool(driver_info, cmd)
        except (exception.PasswordFileFailedToCreate,
                processutils.ProcessExecutionError) as e:
            LOG.warning('IPMI set boot device failed for node %(node)s '
                        'when executing "ipmitool %(cmd)s". '
                        'Error: %(error)s',
                        {'node': driver_info['uuid'], 'cmd': cmd, 'error': e})
            raise exception.IPMIFailure(cmd=cmd)

    @METRICS.timer('IPMIManagement.get_boot_device')
    def get_boot_device(self, task):
        """Get the current boot device for the task's node.

        Returns the current boot device of the node.

        :param task: a task from TaskManager.
        :raises: InvalidParameterValue if required IPMI parameters
            are missing.
        :raises: IPMIFailure on an error from ipmitool.
        :raises: MissingParameterValue if a required parameter is missing.
        :returns: a dictionary containing:

            :boot_device: the boot device, one of
                :mod:`ironic.common.boot_devices` or None if it is unknown.
            :persistent: Whether the boot device will persist to all
                future boots or not, None if it is unknown.

        """
        driver_info = task.node.driver_info
        node = task.node
        ifbd = driver_info.get('ipmi_force_boot_device', False)

        driver_info = _parse_driver_info(task.node)

        if (strutils.bool_from_string(ifbd)
                and node.driver_internal_info.get('persistent_boot_device')
                and node.driver_internal_info.get('is_next_boot_persistent',
                                                  True)):
            return {
                'boot_device': node.driver_internal_info[
                    'persistent_boot_device'],
                'persistent': True
            }

        cmd = "chassis bootparam get 5"
        response = {'boot_device': None, 'persistent': None}

        try:
            out, err = _exec_ipmitool(driver_info, cmd)
        except (exception.PasswordFileFailedToCreate,
                processutils.ProcessExecutionError) as e:
            LOG.warning('IPMI get boot device failed for node %(node)s '
                        'when executing "ipmitool %(cmd)s". '
                        'Error: %(error)s',
                        {'node': driver_info['uuid'], 'cmd': cmd, 'error': e})
            raise exception.IPMIFailure(cmd=cmd)
        # FIXME(TheJulia): This whole thing needs to be refactored to be based
        # upon the response generated by the "Boot parameter data".
        # See: https://storyboard.openstack.org/#!/story/2008241
        re_obj = re.search('Boot Device Selector : (.+)?\n', out)
        if re_obj:
            boot_selector = re_obj.groups('')[0]
            if 'PXE' in boot_selector:
                response['boot_device'] = boot_devices.PXE
            elif 'Hard-Drive' in boot_selector:
                if 'Safe-Mode' in boot_selector:
                    response['boot_device'] = boot_devices.SAFE
                else:
                    response['boot_device'] = boot_devices.DISK
            elif 'BIOS' in boot_selector:
                response['boot_device'] = boot_devices.BIOS
            elif 'CD/DVD' in boot_selector:
                response['boot_device'] = boot_devices.CDROM

        response['persistent'] = 'Options apply to all future boots' in out
        return response

    @METRICS.timer('IPMIManagement.detect_vendor')
    def detect_vendor(self, task):
        """Detects and returns the hardware vendor.

        :param task: A task from TaskManager.
        :raises: InvalidParameterValue if an invalid component, indicator
            or state is specified.
        :raises: MissingParameterValue if a required parameter is missing
        :returns: String representing the BMC reported Vendor or
                  Manufacturer, otherwise returns None.
        """
        driver_info = _parse_driver_info(task.node)
        out, err = _exec_ipmitool(driver_info, "mc info")
        re_obj = re.search("Manufacturer Name *: (.+)", out)
        if re_obj:
            return re_obj.group(1).lower()

    @METRICS.timer('IPMIManagement.get_sensors_data')
    def get_sensors_data(self, task):
        """Get sensors data.

        :param task: a TaskManager instance.
        :raises: FailedToGetSensorData when getting the sensor data fails.
        :raises: FailedToParseSensorData when parsing sensor data fails.
        :raises: InvalidParameterValue if required ipmi parameters are missing
        :raises: MissingParameterValue if a required parameter is missing.
        :returns: returns a dict of sensor data group by sensor type.

        """
        driver_info = _parse_driver_info(task.node)
        # with '-v' option, we can get the entire sensor data including the
        # extended sensor information
        cmd = "sdr -v"
        try:
            out, err = _exec_ipmitool(
                driver_info, cmd, kill_on_timeout=CONF.ipmi.kill_on_timeout)
        except (exception.PasswordFileFailedToCreate,
                processutils.ProcessExecutionError) as e:
            raise exception.FailedToGetSensorData(node=task.node.uuid,
                                                  error=e)

        return _parse_ipmi_sensors_data(task.node, out)

    @METRICS.timer('IPMIManagement.inject_nmi')
    @task_manager.require_exclusive_lock
    def inject_nmi(self, task):
        """Inject NMI, Non Maskable Interrupt.

        Inject NMI (Non Maskable Interrupt) for a node immediately.

        :param task: A TaskManager instance containing the node to act on.
        :raises: IPMIFailure on an error from ipmitool.
        :returns: None

        """
        driver_info = _parse_driver_info(task.node)
        try:
            _exec_ipmitool(driver_info, "power diag")
        except (exception.PasswordFileFailedToCreate,
                processutils.ProcessExecutionError) as err:
            LOG.error('Inject NMI failed for node %(node)s: %(err)s.',
                      {'node': task.node.uuid, 'err': err})
            raise exception.IPMIFailure(cmd="power diag")


class VendorPassthru(base.VendorInterface):

    def __init__(self):
        _constructor_checks(driver=self.__class__.__name__)

    _send_raw_step_args = {
        'raw_bytes': {
            'description': (
                'A string of raw bytes to convey the request to transmit '
                'to the remote BMC.'
            ),
            'required': True
        }
    }

    @METRICS.timer('VendorPassthru.send_raw')
    @base.service_step(priority=0,
                       argsinfo=_send_raw_step_args)
    @base.deploy_step(priority=0,
                      argsinfo=_send_raw_step_args)
    @base.clean_step(priority=0, abortable=False,
                     argsinfo=_send_raw_step_args,
                     requires_ramdisk=False)
    @base.passthru(['POST'],
                   description=_("Send raw bytes to the BMC. Required "
                                 "argument: 'raw_bytes' - a string of raw "
                                 "bytes (e.g. '0x00 0x01')."))
    @task_manager.require_exclusive_lock
    def send_raw(self, task, **kwargs):
        """Send raw bytes to the BMC. Bytes should be a string of bytes.

        :param task: a TaskManager instance.
        :param raw_bytes: a string of raw bytes to send, e.g. '0x00 0x01'
                          supplied as a kwargument.
        :raises: IPMIFailure on an error from ipmitool.
        :raises: MissingParameterValue if a required parameter is missing.
        :raises:  InvalidParameterValue when an invalid value is specified.

        """
        raw_bytes = kwargs.get('raw_bytes')
        if not raw_bytes:
            raise exception.MissingParameterValue(
                'Argument raw_bytes is missing.')
        send_raw(task, raw_bytes)

    @METRICS.timer('VendorPassthru.bmc_reset')
    @base.passthru(['POST'],
                   description=_("Reset the BMC. Required argument: 'warm' "
                                 "(Boolean) - for warm (True) or cold (False) "
                                 "reset."))
    @task_manager.require_exclusive_lock
    def bmc_reset(self, task, http_method, warm=True):
        """Reset BMC with IPMI command 'bmc reset (warm|cold)'.

        :param task: a TaskManager instance.
        :param http_method: the HTTP method used on the request.
        :param warm: boolean parameter to decide on warm or cold reset.
        :raises: IPMIFailure on an error from ipmitool.
        :raises: MissingParameterValue if a required parameter is missing.
        :raises: InvalidParameterValue when an invalid value is specified

        """
        node_uuid = task.node.uuid

        warm = strutils.bool_from_string(warm)
        if warm:
            warm_param = 'warm'
        else:
            warm_param = 'cold'

        LOG.debug('Doing %(warm)s BMC reset on node %(node)s',
                  {'warm': warm_param, 'node': node_uuid})
        driver_info = _parse_driver_info(task.node)
        cmd = 'bmc reset %s' % warm_param

        try:
            out, err = _exec_ipmitool(driver_info, cmd)
            LOG.debug('bmc reset returned stdout: %(stdout)s, stderr:'
                      ' %(stderr)s', {'stdout': out, 'stderr': err})
        except (exception.PasswordFileFailedToCreate,
                processutils.ProcessExecutionError) as e:
            LOG.exception('IPMI "bmc reset" failed for node %(node_id)s '
                          'with error: %(error)s.',
                          {'node_id': node_uuid, 'error': e})
            raise exception.IPMIFailure(cmd=cmd)

    def get_properties(self):
        return COMMON_PROPERTIES

    @METRICS.timer('VendorPassthru.validate')
    def validate(self, task, method, **kwargs):
        """Validate vendor-specific actions.

        If invalid, raises an exception; otherwise returns None.

        Valid methods:
          * send_raw
          * bmc_reset

        :param task: a task from TaskManager.
        :param method: method to be validated
        :param kwargs: info for action.
        :raises: InvalidParameterValue when an invalid parameter value is
                 specified.
        :raises: MissingParameterValue if a required parameter is missing.

        """
        if method == 'send_raw':
            if not kwargs.get('raw_bytes'):
                raise exception.MissingParameterValue(_(
                    'Parameter raw_bytes (string of bytes) was not '
                    'specified.'))

        _parse_driver_info(task.node)


class IPMIConsole(base.ConsoleInterface):
    """A base ConsoleInterface that uses ipmitool."""

    def __init__(self):
        _constructor_checks(driver=self.__class__.__name__)

    def get_properties(self):
        d = COMMON_PROPERTIES.copy()
        d.update(CONSOLE_PROPERTIES)
        return d

    @METRICS.timer('IPMIConsole.validate')
    def validate(self, task):
        """Validate the Node console info.

        :param task: a task from TaskManager.
        :raises: InvalidParameterValue
        :raises: MissingParameterValue when a required parameter is missing

        """
        driver_info = _parse_driver_info(task.node)
        if not driver_info['port'] and CONF.console.port_range is None:
            raise exception.MissingParameterValue(_(
                "Either missing 'ipmi_terminal_port' parameter in node's "
                "driver_info or [console]port_range is not configured"))

        if driver_info['protocol_version'] != '2.0':
            raise exception.InvalidParameterValue(_(
                "Serial over lan only works with IPMI protocol version 2.0. "
                "Check the 'ipmi_protocol_version' parameter in "
                "node's driver_info"))

    def _get_ipmi_cmd(self, driver_info, pw_file):
        """Get ipmi command for ipmitool usage.

        :param driver_info: driver info with the ipmitool parameters
        :param pw_file: password file to be used in ipmitool command
        :returns: returns a command string for ipmitool
        """
        return ' '.join(_get_ipmitool_args(driver_info, pw_file=pw_file))

    def _start_console(self, driver_info, start_method):
        """Start a remote console for the node.

        :param driver_info: the parameters for accessing a node
        :param start_method: console_utils method to start console
        :raises: InvalidParameterValue if required ipmi parameters are missing
        :raises: PasswordFileFailedToCreate if unable to create a file
                 containing the password
        :raises: ConsoleError if the directory for the PID file cannot be
                 created
        :raises: ConsoleSubprocessFailed when invoking the subprocess failed
        """
        try:
            cmd = self._get_ipmi_cmd(driver_info, pw_file=None)
            cmd += ' sol activate'
            start_method(driver_info['uuid'], driver_info['port'], cmd)
        except (exception.ConsoleError,
                exception.ConsoleSubprocessFailed) as e:
            LOG.exception('IPMI Error while attempting "%(cmd)s" '
                          'for node %(node)s. Error: %(error)s',
                          {'node': driver_info['uuid'],
                           'cmd': cmd, 'error': e})
            raise


class IPMIShellinaboxConsole(IPMIConsole):
    """A ConsoleInterface that uses ipmitool and shellinabox."""

    # TODO(TheJulia): This interface is deprecated due to the shellinabox
    # project being abandoned. It should be removed after the release of
    # 2025.2 in advance of 2026.1's release.
    # https://github.com/shellinabox/shellinabox
    # https://github.com/shellinabox/shellinabox/issues/531

    supported = False

    def _get_ipmi_cmd(self, driver_info, pw_file):
        """Get ipmi command for ipmitool usage.

        :param driver_info: driver info with the ipmitool parameters
        :param pw_file: password file to be used in ipmitool command
        :returns: returns a command string for ipmitool
        """
        command = super(IPMIShellinaboxConsole, self)._get_ipmi_cmd(
            driver_info, pw_file)
        return ("/:%(uid)s:%(gid)s:HOME:%(basic_command)s"
                % {'uid': os.getuid(),
                   'gid': os.getgid(),
                   'basic_command': command})

    @METRICS.timer('IPMIShellinaboxConsole.start_console')
    def start_console(self, task):
        """Start a remote console for the node.

        :param task: a task from TaskManager
        :raises: InvalidParameterValue if required ipmi parameters are missing
        :raises: PasswordFileFailedToCreate if unable to create a file
                 containing the password
        :raises: ConsoleError if the directory for the PID file cannot be
                 created
        :raises: ConsoleSubprocessFailed when invoking the subprocess failed
        """
        # Dealloc allocated port if any, so the same host can never has
        # duplicated port.
        _release_allocated_port(task)
        driver_info = _parse_driver_info(task.node)
        if not driver_info['port']:
            driver_info['port'] = _allocate_port(task)

        try:
            self._exec_stop_console(driver_info)
        except OSError:
            # We need to drop any existing sol sessions with sol deactivate.
            # OSError is raised when sol session is already deactivated,
            # so we can ignore it.
            pass
        self._start_console(driver_info,
                            console_utils.start_shellinabox_console)

    @METRICS.timer('IPMIShellinaboxConsole.stop_console')
    def stop_console(self, task):
        """Stop the remote console session for the node.

        :param task: a task from TaskManager
        :raises: ConsoleError if unable to stop the console
        """
        try:
            console_utils.stop_shellinabox_console(task.node.uuid)
        finally:
            utils.unlink_without_raise(_console_pwfile_path(task.node.uuid))
        _release_allocated_port(task)

    def _exec_stop_console(self, driver_info):
        cmd = "sol deactivate"
        _exec_ipmitool(driver_info, cmd, check_exit_code=[0, 1])

    @METRICS.timer('IPMIShellinaboxConsole.get_console')
    def get_console(self, task):
        """Get the type and connection information about the console."""
        driver_info = _parse_driver_info(task.node)

        try:
            self._exec_stop_console(driver_info)
        except OSError:
            # We need to drop any existing sol sessions with sol deactivate.
            # OSError is raised when sol session is already deactivated,
            # so we can ignore it.
            pass
        url = console_utils.get_shellinabox_console_url(driver_info['port'])
        return {'type': 'shellinabox', 'url': url}


class IPMISocatConsole(IPMIConsole):
    """A ConsoleInterface that uses ipmitool and socat."""

    @METRICS.timer('IPMISocatConsole.start_console')
    def start_console(self, task):
        """Start a remote console for the node.

        :param task: a task from TaskManager
        :raises: InvalidParameterValue if required ipmi parameters are missing
        :raises: PasswordFileFailedToCreate if unable to create a file
                 containing the password
        :raises: ConsoleError if the directory for the PID file cannot be
                 created
        :raises: ConsoleSubprocessFailed when invoking the subprocess failed
        """
        # Dealloc allocated port if any, so the same host can never has
        # duplicated port.
        _release_allocated_port(task)
        driver_info = _parse_driver_info(task.node)
        if not driver_info['port']:
            driver_info['port'] = _allocate_port(
                task, host=CONF.console.socat_address)

        try:
            self._exec_stop_console(driver_info)
        except OSError:
            # We need to drop any existing sol sessions with sol deactivate.
            # OSError is raised when sol session is already deactivated,
            # so we can ignore it.
            pass
        self._start_console(driver_info, console_utils.start_socat_console)

    @METRICS.timer('IPMISocatConsole.stop_console')
    def stop_console(self, task):
        """Stop the remote console session for the node.

        :param task: a task from TaskManager
        :raises: ConsoleError if unable to stop the console
        """
        driver_info = _parse_driver_info(task.node)
        try:
            console_utils.stop_socat_console(task.node.uuid)
        finally:
            utils.unlink_without_raise(_console_pwfile_path(task.node.uuid))
        self._exec_stop_console(driver_info)
        _release_allocated_port(task)

    def _exec_stop_console(self, driver_info):
        cmd = "sol deactivate"
        _exec_ipmitool(driver_info, cmd, check_exit_code=[0, 1])

    @METRICS.timer('IPMISocatConsole.get_console')
    def get_console(self, task):
        """Get the type and connection information about the console.

        :param task: a task from TaskManager
        """
        driver_info = _parse_driver_info(task.node)

        try:
            self._exec_stop_console(driver_info)
        except OSError:
            # We need to drop any existing sol sessions with sol deactivate.
            # OSError is raised when sol session is already deactivated,
            # so we can ignore it.
            pass
        url = console_utils.get_socat_console_url(driver_info['port'])
        return {'type': 'socat', 'url': url}



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Common functionalities shared between different iRMC modules.
"""
import json
import os
import re

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import importutils
from oslo_utils import strutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.conf import CONF
from ironic.drivers.modules import snmp

scci = importutils.try_import('scciclient.irmc.scci')
elcm = importutils.try_import('scciclient.irmc.elcm')

LOG = logging.getLogger(__name__)


IRMC_OS_NAME_R = re.compile(r'iRMC\s+S\d+')
IRMC_OS_NAME_NUM_R = re.compile(r'\d+$')
IRMC_FW_VER_R = re.compile(r'\d(\.\d+)*\w*')
IRMC_FW_VER_NUM_R = re.compile(r'\d(\.\d+)*')

IPMI_ENABLED_BY_DEFAULT_RANGES = {
    # iRMC S4 enables IPMI over LAN by default
    '4': None,
    # iRMC S5 enables IPMI over LAN by default
    '5': None,
    # iRMC S6 disables IPMI over LAN by default from version 2.00
    '6': {'upper': '2.00'}}

ELCM_STATUS_PATH = '/rest/v1/Oem/eLCM/eLCMStatus'

# List of xxx_interface & implementation pair which uses SNMP internally
# and iRMC driver supports
INTERFACE_IMPL_LIST_WITH_SNMP = {
    'inspect_interface': {'irmc', },
    'power_interface': {'irmc', }}

REQUIRED_PROPERTIES = {
    'irmc_address': _("IP address or hostname of the iRMC. Required."),
    'irmc_username': _("Username for the iRMC with administrator privileges. "
                       "Required."),
    'irmc_password': _("Password for irmc_username. Required."),
}
OPTIONAL_PROPERTIES = {
    'irmc_port': _("Port to be used for iRMC operations; either 80 or 443. "
                   "The default value is 443. Optional."),
    'irmc_auth_method': _("Authentication method for iRMC operations; "
                          "either 'basic' or 'digest'. The default value is "
                          "'basic'. Optional."),
    'irmc_client_timeout': _("Timeout (in seconds) for iRMC operations. "
                             "The default value is 60. Optional."),
    'irmc_sensor_method': _("Sensor data retrieval method; either "
                            "'ipmitool' or 'scci'. The default value is "
                            "'ipmitool'. Optional."),
}
OPTIONAL_DRIVER_INFO_PROPERTIES = {
    'irmc_verify_ca': _('Either a Boolean value, a path to a CA_BUNDLE '
                        'file or directory with certificates of trusted '
                        'CAs. If set to True the driver will verify the '
                        'host certificates; if False the driver will '
                        'ignore verifying the SSL certificate. If it\'s '
                        'a path the driver will use the specified '
                        'certificate or one of the certificates in the '
                        'directory. Defaults to True. Optional'),
}

SNMP_PROPERTIES = {
    'irmc_snmp_version': _("SNMP protocol version; either 'v1', 'v2c', or "
                           "'v3'. The default value is 'v2c'. Optional."),
    'irmc_snmp_port': _("SNMP port. The default is 161. Optional."),
    'irmc_snmp_community': _("SNMP community required for versions 'v1' and "
                             "'v2c'. The default value is 'public'. "
                             "Optional."),
}

SNMP_V3_REQUIRED_PROPERTIES = {
    'irmc_snmp_user': _("SNMPv3 User-based Security Model (USM) username. "
                        "Required for version 'v3’. "),
    'irmc_snmp_auth_password': _("SNMPv3 message authentication key. Must be "
                                 "8+ characters long. Required when message "
                                 "authentication is used."),
    'irmc_snmp_priv_password': _("SNMPv3 message privacy key. Must be 8+ "
                                 "characters long. Required when message "
                                 "privacy is used."),
}

SNMP_V3_OPTIONAL_PROPERTIES = {
    'irmc_snmp_auth_proto': _("SNMPv3 message authentication protocol ID. "
                              "Required for version 'v3'. "
                              "If using iRMC S4/S5, only 'sha' is supported."
                              "If using iRMC S6, the valid options are "
                              "'sha256', 'sha384', 'sha512'."),
    'irmc_snmp_priv_proto': _("SNMPv3 message privacy (encryption) protocol "
                              "ID. Required for version 'v3'. "
                              "'aes' is supported."),
}

SNMP_V3_DEPRECATED_PROPERTIES = {
    'irmc_snmp_security': _("SNMP security name required for version 'v3'. "
                            "Optional. Deprecated."),
}


COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)
COMMON_PROPERTIES.update(OPTIONAL_DRIVER_INFO_PROPERTIES)
COMMON_PROPERTIES.update(SNMP_PROPERTIES)
COMMON_PROPERTIES.update(SNMP_V3_REQUIRED_PROPERTIES)
COMMON_PROPERTIES.update(SNMP_V3_OPTIONAL_PROPERTIES)
COMMON_PROPERTIES.update(SNMP_V3_DEPRECATED_PROPERTIES)


def parse_driver_info(node):
    """Gets the specific Node driver info.

    This method validates whether the 'driver_info' property of the
    supplied node contains the required information for this driver.

    :param node: An ironic node object.
    :returns: A dict containing information from driver_info
        and default values.
    :raises: InvalidParameterValue if invalid value is contained
        in the 'driver_info' property.
    :raises: MissingParameterValue if some mandatory key is missing
        in the 'driver_info' property.
    """
    info = node.driver_info
    missing_info = [key for key in REQUIRED_PROPERTIES if not info.get(key)]
    if missing_info:
        raise exception.MissingParameterValue(_(
            "Missing the following iRMC parameters in node's"
            " driver_info: %s.") % missing_info)

    req = {key: value for key, value in info.items()
           if key in REQUIRED_PROPERTIES}
    # corresponding config names don't have 'irmc_' prefix
    opt = {param: info.get(param, CONF.irmc.get(param[len('irmc_'):]))
           for param in OPTIONAL_PROPERTIES}
    opt_driver_info = {param: info.get(param)
                       for param in OPTIONAL_DRIVER_INFO_PROPERTIES}
    d_info = dict(req, **opt, **opt_driver_info)
    d_info['irmc_port'] = utils.validate_network_port(
        d_info['irmc_port'], 'irmc_port')

    error_msgs = []
    if (d_info['irmc_auth_method'].lower() not in ('basic', 'digest')):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_auth_method'.") %
            d_info['irmc_auth_method'])
    if d_info['irmc_port'] not in (80, 443):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_port'.") %
            d_info['irmc_port'])
    if not isinstance(d_info['irmc_client_timeout'], int):
        error_msgs.append(
            _("Value '%s' is not an integer for 'irmc_client_timeout'") %
            d_info['irmc_client_timeout'])
    if d_info['irmc_sensor_method'].lower() not in ('ipmitool', 'scci'):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_sensor_method'.") %
            d_info['irmc_sensor_method'])

    verify_ca = d_info.get('irmc_verify_ca')
    if verify_ca is None:
        d_info['irmc_verify_ca'] = verify_ca = CONF.webserver_verify_ca

    # Check if verify_ca is a Boolean or a file/directory in the file-system
    if isinstance(verify_ca, str):
        if ((os.path.isdir(verify_ca) and os.path.isabs(verify_ca))
            or (os.path.isfile(verify_ca) and os.path.isabs(verify_ca))):
            # If it's fullpath and dir/file, we don't need to do anything
            pass
        else:
            try:
                d_info['irmc_verify_ca'] = strutils.bool_from_string(
                    verify_ca, strict=True)
            except ValueError:
                error_msgs.append(
                    _('Invalid value type set in driver_info/'
                      'irmc_verify_ca on node %(node)s. '
                      'The value should be a Boolean or the path '
                      'to a file/directory, not "%(value)s"'
                      ) % {'value': verify_ca, 'node': node.uuid})
    elif isinstance(verify_ca, bool):
        # If it's a boolean it's grand, we don't need to do anything
        pass
    else:
        error_msgs.append(
            _('Invalid value type set in driver_info/irmc_verify_ca '
              'on node %(node)s. The value should be a Boolean or the path '
              'to a file/directory, not "%(value)s"') % {'value': verify_ca,
                                                         'node': node.uuid})

    if error_msgs:
        msg = (_("The following errors were encountered while parsing "
                 "driver_info:\n%s") % "\n".join(error_msgs))
        raise exception.InvalidParameterValue(msg)

    d_info.update(_parse_snmp_driver_info(node, info))

    return d_info


def _parse_snmp_driver_info(node, info):
    """Parses the SNMP related driver_info parameters.

    :param node: An Ironic node object.
    :param info: driver_info dictionary.
    :returns: A dictionary containing SNMP information.
    :raises: MissingParameterValue if any of the mandatory
        parameter values are not provided.
    :raises: InvalidParameterValue if there is any invalid
        value provided.
    """
    snmp_info = {param: info.get(param, CONF.irmc.get(param[len('irmc_'):]))
                 for param in SNMP_PROPERTIES}
    valid_versions = {"v1": snmp.SNMP_V1,
                      "v2c": snmp.SNMP_V2C,
                      "v3": snmp.SNMP_V3}

    for int_name, impl_list in INTERFACE_IMPL_LIST_WITH_SNMP.items():
        if getattr(node, int_name) in impl_list:
            break
    else:
        return snmp_info

    if snmp_info['irmc_snmp_version'].lower() not in valid_versions:
        raise exception.InvalidParameterValue(_(
            "Value '%s' is not supported for 'irmc_snmp_version'.") %
            snmp_info['irmc_snmp_version']
        )
    snmp_info["irmc_snmp_version"] = \
        valid_versions[snmp_info["irmc_snmp_version"].lower()]

    snmp_info['irmc_snmp_port'] = utils.validate_network_port(
        snmp_info['irmc_snmp_port'], 'irmc_snmp_port')

    if snmp_info['irmc_snmp_version'] != snmp.SNMP_V3:
        if (snmp_info['irmc_snmp_community']
            and not isinstance(snmp_info['irmc_snmp_community'], str)):
            raise exception.InvalidParameterValue(_(
                "Value '%s' is not a string for 'irmc_snmp_community'") %
                snmp_info['irmc_snmp_community'])
        if utils.is_fips_enabled():
            raise exception.InvalidParameterValue(_(
                "'v3' has to be set for 'irmc_snmp_version' "
                "when FIPS mode is enabled."))

    else:
        snmp_info.update(_parse_snmp_v3_info(node, info))

    return snmp_info


def _parse_snmp_v3_info(node, info):
    snmp_info = {}
    missing_info = []
    valid_values = {'irmc_snmp_auth_proto': ['sha', 'sha256', 'sha384',
                                             'sha512'],
                    'irmc_snmp_priv_proto': ['aes']}
    valid_protocols = {'irmc_snmp_auth_proto': snmp.snmp_auth_protocols,
                       'irmc_snmp_priv_proto': snmp.snmp_priv_protocols}
    snmp_keys = {'irmc_snmp_auth_password', 'irmc_snmp_priv_password'}

    security = info.get('irmc_snmp_security', CONF.irmc.get('snmp_security'))
    for param in SNMP_V3_REQUIRED_PROPERTIES:
        try:
            snmp_info[param] = info[param]
        except KeyError:
            if param == 'irmc_snmp_user':
                if not security:
                    missing_info.append(param)
                else:
                    LOG.warning(_("'irmc_snmp_security' parameter is "
                                  "deprecated in favor of 'irmc_snmp_user' "
                                  "parameter. Please set 'irmc_snmp_user' "
                                  "and remove 'irmc_snmp_security' for node "
                                  "%s."), node.uuid)
                    # In iRMC, the username must start with a letter, so only
                    # a string can be a valid username and a string from a
                    # number is invalid.
                    if not isinstance(security, str):
                        raise exception.InvalidParameterValue(_(
                            "Value '%s' is not a string for "
                            "'irmc_snmp_security.") %
                            info['irmc_snmp_security'])
                    else:
                        snmp_info['irmc_snmp_user'] = security
                        security = None
            else:
                missing_info.append(param)

    if missing_info:
        raise exception.MissingParameterValue(_(
            "The following required SNMP parameters "
            "are missing: %s") % missing_info)

    if security:
        LOG.warning(_("'irmc_snmp_security' parameter is ignored in favor of "
                      "'irmc_snmp_user' parameter. Please remove "
                      "'irmc_snmp_security' from node %s "
                      "configuration."), node.uuid)
    if not isinstance(snmp_info['irmc_snmp_user'], str):
        raise exception.InvalidParameterValue(_(
            "Value '%s' is not a string for 'irmc_snmp_user'.") %
            info['irmc_snmp_user'])

    for param in snmp_keys:
        if not isinstance(snmp_info[param], str):
            raise exception.InvalidParameterValue(_(
                "Value %(value)s is not a string for %(param)s.") %
                {'param': param, 'value': snmp_info[param]})
        if len(snmp_info[param]) < 8:
            raise exception.InvalidParameterValue(_(
                "%s is too short. (8+ chars required)") % param)

    for param in SNMP_V3_OPTIONAL_PROPERTIES:
        value = None
        try:
            value = info[param]
            if value not in valid_values[param]:
                raise exception.InvalidParameterValue(_(
                    "Invalid value %(value)s given for driver info parameter "
                    "%(param)s, the valid values are %(valid_values)s.") %
                    {'param': param,
                     'value': value,
                     'valid_values': valid_values[param]})
        except KeyError:
            value = CONF.irmc.get(param[len('irmc_'):])
        snmp_info[param] = valid_protocols[param].get(value)
        if not snmp_info[param]:
            raise exception.InvalidParameterValue(_(
                "Unknown SNMPv3 protocol %(value)s given for "
                "driver info parameter %(param)s") % {'param': param,
                                                      'value': value})

    return snmp_info


def get_irmc_client(node):
    """Gets an iRMC SCCI client.

    Given an ironic node object, this method gives back a iRMC SCCI client
    to do operations on the iRMC.

    :param node: An ironic node object.
    :returns: scci_cmd partial function which takes a SCCI command param.
    :raises: InvalidParameterValue on invalid inputs.
    :raises: MissingParameterValue if some mandatory information
        is missing on the node
    :raises: IRMCOperationError if iRMC operation failed
    """
    driver_info = parse_driver_info(node)

    scci_client = scci.get_client(
        driver_info['irmc_address'],
        driver_info['irmc_username'],
        driver_info['irmc_password'],
        port=driver_info['irmc_port'],
        auth_method=driver_info['irmc_auth_method'],
        verify=driver_info.get('irmc_verify_ca'),
        client_timeout=driver_info['irmc_client_timeout'])
    return scci_client


def update_ipmi_properties(task):
    """Update ipmi properties to node driver_info.

    :param task: A task from TaskManager.
    """
    node = task.node
    info = node.driver_info

    # updating ipmi credentials
    info['ipmi_address'] = info.get('irmc_address')
    info['ipmi_username'] = info.get('irmc_username')
    info['ipmi_password'] = info.get('irmc_password')

    # saving ipmi credentials to task object
    task.node.driver_info = info


def get_irmc_report(node):
    """Gets iRMC SCCI report.

    Given an ironic node object, this method gives back a iRMC SCCI report.

    :param node: An ironic node object.
    :returns: A xml.etree.ElementTree object.
    :raises: InvalidParameterValue on invalid inputs.
    :raises: MissingParameterValue if some mandatory information
        is missing on the node.
    :raises: scci.SCCIInvalidInputError if required parameters are invalid.
    :raises: scci.SCCIClientError if SCCI failed.
    """
    driver_info = parse_driver_info(node)

    return scci.get_report(
        driver_info['irmc_address'],
        driver_info['irmc_username'],
        driver_info['irmc_password'],
        port=driver_info['irmc_port'],
        auth_method=driver_info['irmc_auth_method'],
        verify=driver_info.get('irmc_verify_ca'),
        client_timeout=driver_info['irmc_client_timeout'])


def get_secure_boot_mode(node):
    """Get the current secure boot mode.

    :param node: An ironic node object.
    :raises: UnsupportedDriverExtension if secure boot is not present.
    :raises: IRMCOperationError if the operation fails.
    """
    driver_info = parse_driver_info(node)

    try:
        return elcm.get_secure_boot_mode(driver_info)
    except elcm.SecureBootConfigNotFound:
        raise exception.UnsupportedDriverExtension(
            driver=node.driver, extension='get_secure_boot_state')
    except scci.SCCIError as irmc_exception:
        LOG.error("Failed to get secure boot for node %s", node.uuid)
        raise exception.IRMCOperationError(
            operation=_("getting secure boot mode"),
            error=irmc_exception)


def set_secure_boot_mode(node, enable):
    """Enable or disable UEFI Secure Boot

    :param node: An ironic node object.
    :param enable: Boolean value. True if the secure boot to be
        enabled.
    :raises: IRMCOperationError if the operation fails.
    """
    driver_info = parse_driver_info(node)

    try:
        elcm.set_secure_boot_mode(driver_info, enable)
        LOG.info("Set secure boot to %(flag)s for node %(node)s",
                 {'flag': enable, 'node': node.uuid})
    except scci.SCCIError as irmc_exception:
        LOG.error("Failed to set secure boot to %(flag)s for node %(node)s",
                  {'flag': enable, 'node': node.uuid})
        raise exception.IRMCOperationError(
            operation=_("setting secure boot mode"),
            error=irmc_exception)


def check_elcm_license(node):
    """Connect to iRMC and return status of eLCM license

    This function connects to iRMC REST API and check whether eLCM
    license is active. This function can be used to check connection to
    iRMC REST API.

    :param node: An ironic node object
    :returns: dictionary whose keys are 'active' and 'status_code'.
        value of 'active' is boolean showing if eLCM license is active
        and value of 'status_code' is int which is HTTP return code
        from iRMC REST API access
    :raises: InvalidParameterValue if invalid value is contained
        in the 'driver_info' property.
    :raises: MissingParameterValue if some mandatory key is missing
        in the 'driver_info' property.
    :raises: IRMCOperationError if the operation fails.
    """
    try:
        d_info = parse_driver_info(node)
        # GET to /rest/v1/Oem/eLCM/eLCMStatus returns
        # JSON data like this:
        #
        # {
        # "eLCMStatus":{
        #      "EnabledAndLicenced":"true",
        #      "SDCardMounted":"false"
        #      }
        # }
        #
        # EnabledAndLicenced tells whether eLCM license is valid
        #
        r = elcm.elcm_request(d_info, 'GET', ELCM_STATUS_PATH)

        # If r.status_code is 200, it means success and r.text is JSON.
        # If it is 500, it means there is problem at iRMC side
        # and iRMC cannot return eLCM status.
        # If it was 401, elcm_request raises SCCIClientError.
        # Otherwise, r.text may not be JSON.
        if r.status_code == 200:
            license_active = strutils.bool_from_string(
                jsonutils.loads(r.text)['eLCMStatus']['EnabledAndLicenced'],
                strict=True)
        else:
            license_active = False

        return {'active': license_active, 'status_code': r.status_code}
    except (scci.SCCIError,
            json.JSONDecodeError,
            TypeError,
            KeyError,
            ValueError) as irmc_exception:
        LOG.error("Failed to check eLCM license status for node $(node)s",
                  {'node': node.uuid})
        raise exception.IRMCOperationError(
            operation='checking eLCM license status',
            error=irmc_exception)


def set_irmc_version(task):
    """Fetch and save iRMC firmware version.

    This function should be called before calling any other functions which
    need to check node's iRMC firmware version.

    Set `<iRMC OS>/<fw version>` to driver_internal_info['irmc_fw_version']

    :param node: An ironic node object
    :raises: InvalidParameterValue if invalid value is contained
        in the 'driver_info' property.
    :raises: MissingParameterValue if some mandatory key is missing
        in the 'driver_info' property.
    :raises: IRMCOperationError if the operation fails.
    :raises: NodeLocked if the target node is already locked.
    """

    node = task.node
    try:
        report = get_irmc_report(node)
        irmc_os, fw_version = scci.get_irmc_version_str(report)

        fw_ver = node.driver_internal_info.get('irmc_fw_version')
        if fw_ver != '/'.join([irmc_os, fw_version]):
            task.upgrade_lock(purpose='saving firmware version')
            node.set_driver_internal_info('irmc_fw_version',
                                          f"{irmc_os}/{fw_version}")
            node.save()
    except scci.SCCIError as irmc_exception:
        LOG.error("Failed to fetch iRMC FW version for node %s",
                  node.uuid)
        raise exception.IRMCOperationError(
            operation=_("fetching irmc fw version "),
            error=irmc_exception)


def _version_lt(v1, v2):
    v1_l = v1.split('.')
    v2_l = v2.split('.')
    if len(v1_l) <= len(v2_l):
        v1_l.extend(['0'] * (len(v2_l) - len(v1_l)))
    else:
        v2_l.extend(['0'] * (len(v1_l) - len(v2_l)))

    for i in range(len(v1_l)):
        if int(v1_l[i]) < int(v2_l[i]):
            return True
        elif int(v1_l[i]) > int(v2_l[i]):
            return False
    else:
        return False


def _version_le(v1, v2):
    v1_l = v1.split('.')
    v2_l = v2.split('.')
    if len(v1_l) <= len(v2_l):
        v1_l.extend(['0'] * (len(v2_l) - len(v1_l)))
    else:
        v2_l.extend(['0'] * (len(v1_l) - len(v2_l)))

    for i in range(len(v1_l)):
        if int(v1_l[i]) < int(v2_l[i]):
            return True
        elif int(v1_l[i]) > int(v2_l[i]):
            return False
    else:
        return True


def within_version_ranges(node, version_ranges):
    """Read saved iRMC FW version and check if it is within the passed ranges.

    :param node: An ironic node object
    :param version_ranges: A Python dictionary containing version ranges in the
        next format: <os_n>: <ranges>, where <os_n> is a string representing
        iRMC OS number (e.g. '4') and <ranges> is a dictionaries indicating
        the specific firmware version ranges under the iRMC OS number <os_n>.

        The dictionary used in <ranges> only has two keys: 'min' and 'upper',
        and value of each key is a string representing iRMC firmware version
        number or None. Both keys can be absent and their value can be None.

        It is acceptable to not set ranges for a <os_n> (for example set
        <ranges> to None, {}, etc...), in this case, this function only
        checks if the node's iRMC OS number matches the <os_n>.

        Valid <version_ranges> example:
            {'3': None,   # all version of iRMC S3 matches
             '4': {},     # all version of iRMC S4 matches
             # all version of iRMC S5 matches
             '5': {'min': None, 'upper': None},
             # iRMC S6 whose version is >=1.20 matches
             '6': {'min': '1.20', 'upper': None},
             # iRMC S7 whose version is
             # 5.51<= (version) <8.23 matches
             '7': {'min': '5.51', 'upper': '8.23'}}

    :returns: True if node's iRMC FW is in range, False if not or
        fails to parse firmware version
    """

    try:
        fw_version = node.driver_internal_info.get('irmc_fw_version', '')
        irmc_os, irmc_ver = fw_version.split('/')

        if IRMC_OS_NAME_R.match(irmc_os) and IRMC_FW_VER_R.match(irmc_ver):
            os_num = IRMC_OS_NAME_NUM_R.search(irmc_os).group(0)
            fw_num = IRMC_FW_VER_NUM_R.search(irmc_ver).group(0)

            if os_num not in version_ranges:
                return False

            v_range = version_ranges[os_num]

            # An OS number with no ranges set means no need to check
            # specific version, all the version under this OS number is valid.
            if not v_range:
                return True

            # Specific range is set, check if the node's
            # firmware version is within it.
            min_ver = v_range.get('min')
            upper_ver = v_range.get('upper')
            flag = True
            if min_ver:
                flag = _version_le(min_ver, fw_num)
            if flag and upper_ver:
                flag = _version_lt(fw_num, upper_ver)
            return flag

    except Exception:
        # All exceptions are ignored
        pass

    LOG.warning('Failed to parse iRMC firmware version on node %(uuid)s: '
                '%(fw_ver)s', {'uuid': node.uuid, 'fw_ver': fw_version})
    return False
